unit Aif;

interface

uses
  Windows, CRC;

const
  AIF_FORMAT_EPOC = $10000037;

type
  TAifItem = packed record
    Name: String;
    LangCode: Integer;
  end;

  TAifIcon = class
  private
    FMBMData: Pointer;
    FMBMDataSize: Integer;
    FMBMFileName: String;
    FStrings: array of TAifItem;
    FMimeType: array of TAifItem;
    FAIFFormat: DWORD;

    function GetStrings(Index: Integer): String;
    function GetStringsLangCode(Index: Integer): Integer;
    function GetMimeType(Index: Integer): String;
    function GetMimeTypeLangCode(Index: Integer): Integer;

    procedure SetStrings(Index: Integer; Value: String);
    procedure SetStringsLangCode(Index: Integer; Value: Integer);
    procedure SetMimeType(Index: Integer; Value: String);
    procedure SetMimeTypeLangCode(Index: Integer; Value: Integer);
  public
    // AIF Data
    UIDS: array[1..4] of DWORD;
    Constant: DWORD;
    Embeddable: DWORD;
    NewFile: DWORD;
    Hidden: DWORD;
    UseReserved: Boolean;
    Reserved: array[0..4] of BYTE;
    MIMEID: DWORD;
    MIMEEnd: WORD;

    constructor Create;
    destructor Destroy; override;

    property MBMData: Pointer read FMBMData;
    property MBMDataSize: Integer read FMBMDataSize;
    property MBMFileName: String read FMBMFileName write FMBMFileName;
    property Strings[Index: Integer]: String read GetStrings write SetStrings;
    property StringsLangCode[Index: Integer]: Integer read GetStringsLangCode write SetStringsLangCode;
    property MimeType[Index: Integer]: String read GetMimeType write SetMimeType;
    property MimeTypeLangCode[Index: Integer]: Integer read GetMimeTypeLangCode write SetMimeTypeLangCode;

    property AIFFormat: DWORD read FAIFFormat write FAIFFormat;

    function AddItemStrings: Integer;
    function AddItemMimeType: Integer;

    function CountItemStrings: Integer;
    function CountItemMimeType: Integer;

    procedure DeleteItemStrings(Index: Integer);
    procedure DeleteItemMimeType(Index: Integer);

    function LoadFromFile(FileName: String; MBMToData: Boolean = False): Boolean;
    function SaveToFile(FileName: String; UseMBMData: Boolean = False): Boolean;
  end;

implementation

const
  MBMHeader: array[0..15] of BYTE =
  (
    $37, $00, $00, $10, $42, $00, $00, $10,
    $00, $00, $00, $00, $39, $64, $39, $47
  );

constructor TAifIcon.Create;
begin
  FAIFFormat := AIF_FORMAT_EPOC;
  UIDS[1] := $10000037;
  UIDS[2] := $10003A38;
  UIDS[3] := $00000000;
  UIDS[4] := CalcUID4(UIDS[1], UIDS[2], UIDS[3]);
  Constant := $00000001;
  Embeddable := $00000000;
  NewFile := $00000000;
  Hidden := $00000000;
  MIMEID := $00000003;
  MIMEEnd := $00;
  ZeroMemory(@Reserved, SIZEOF(Reserved));
  UseReserved := False;
end;

destructor TAifIcon.Destroy;
begin
  SetLength(FStrings, 0);
  SetLength(FMimeType, 0);
end;

function TAifIcon.GetStrings(Index: Integer): String;
begin
  Result := FStrings[Index].Name;
end;

function TAifIcon.GetStringsLangCode(Index: Integer): Integer;
begin
  Result := FStrings[Index].LangCode;
end;

function TAifIcon.GetMimeType(Index: Integer): String;
begin
  Result := FMimeType[Index].Name;
end;

function TAifIcon.GetMimeTypeLangCode(Index: Integer): Integer;
begin
  Result := FMimeType[Index].LangCode;
end;

procedure TAifIcon.SetStrings(Index: Integer; Value: String);
begin
  FStrings[Index].Name := Value;  
end;

procedure TAifIcon.SetStringsLangCode(Index: Integer; Value: Integer);
begin
  FStrings[Index].LangCode := Value;
end;

procedure TAifIcon.SetMimeType(Index: Integer; Value: String);
begin
  FMimeType[Index].Name := Value;
end;

procedure TAifIcon.SetMimeTypeLangCode(Index: Integer; Value: Integer);
begin
  FMimeType[Index].LangCode := Value;
end;

function TAifIcon.AddItemStrings: Integer;
begin
  SetLength(FStrings, Length(FStrings) + 1);
  Result := Length(FStrings) - 1;
end;

function TAifIcon.AddItemMimeType: Integer;
begin
  SetLength(FMimeType, Length(FMimeType) + 1);
  Result := Length(FMimeType) - 1;
end;

function TAifIcon.CountItemStrings: Integer;
begin
  Result := Length(FStrings);
end;

function TAifIcon.CountItemMimeType: Integer;
begin
  Result := Length(FMimeType);
end;

procedure TAifIcon.DeleteItemStrings(Index: Integer);
var
  i: Integer;
begin
  for i:= Length(FStrings) - 2 downto Index do
    FStrings[i] := FStrings[i + 1];
  Setlength(FStrings, Length(FStrings) - 1);
end;
                           
procedure TAifIcon.DeleteItemMimeType(Index: Integer);
var
  i: Integer;
begin
  for i:= Length(FMimeType) - 2 downto Index do
    FMimeType[i] := FMimeType[i + 1];
  Setlength(FMimeType, Length(FMimeType) - 1);
end;

function TAifIcon.LoadFromFile;
var
  AIF: HFILE;
  MBM: HFILE;
  lpReOpenBuff: _OFSTRUCT;
  lpNumberOfBytes: DWORD;
  i: Integer;
  Item: Integer;
  Buffer: array[0..1023] of BYTE;
  BufferSize: DWORD;
  TrailerOffset: DWORD;
  StringCount: BYTE;
  StringOffset: DWORD;
  StringLangCode: WORD;
  Cap: PCHAR;
  CapLength: BYTE;
  IconCount: BYTE;
  IconCountInMBM: DWORD;
  IconTrailer: DWORD;
  BitMapOffset: DWORD;
  BitMapSize: WORD;
  MIMETrailer: DWORD;
  MIMECount: BYTE;
  MIMEOffset: DWORD;
  MIMELangCode: WORD;
  MBMDataPosition: Integer;
begin
  Result := False;

  Setlength(FStrings, 0);
  Setlength(FMimeType, 0);
  FAIFFormat := AIF_FORMAT_EPOC;
  UIDS[1] := $10000037;
  UIDS[2] := $10003A38;
  UIDS[3] := $00000000;
  UIDS[4] := CalcUID4(UIDS[1], UIDS[2], UIDS[3]);
  Constant := $00000001;
  Embeddable := $00000000;
  NewFile := $00000000;
  Hidden := $00000000;
  MIMEID := $00000003;
  MIMEEnd := $00;
  ZeroMemory(@Reserved, SIZEOF(Reserved));
  UseReserved := False;

  MIMEID := 0;
  MIMECount := 0;
  MIMEOffset := 0;
  MIMELangCode := 0;
  MIMEEnd := 0;
  UseReserved := False;

  AIF := OpenFile(PCHAR(FileName), lpReOpenBuff, OF_READ);
  if AIF = HFILE_ERROR then
    Exit;

  ReadFile(AIF, UIDS, SIZEOF(UIDS), lpNumberOfBytes, NIL);
                              
  if CalcUID4(UIDS[1], UIDS[2], UIDS[3]) <> UIDS[4] then
  begin
    CloseHandle(AIF);
    Exit;
  end;

  case UIDS[1] of
    AIF_FORMAT_EPOC:
      begin
        FAIFFormat := AIF_FORMAT_EPOC;

        ReadFile(AIF, TrailerOffset, SIZEOF(TrailerOffset), lpNumberOfBytes, NIL);

        // Strings
        SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
        ReadFile(AIF, StringCount, SIZEOF(StringCount), lpNumberOfBytes, NIL);
        StringCount := StringCount div 2;
        for i:= 1 to StringCount do
        begin
          SetFilePointer(
            AIF,
            TrailerOffset + SIZEOF(StringCount) + (SIZEOF(StringOffset) +
            SIZEOF(StringLangCode))*(DWORD(i)-1),
            NIL,
            FILE_BEGIN);
        ReadFile(AIF, StringOffset, SIZEOF(StringOffset), lpNumberOfBytes, NIL);
        ReadFile(AIF, StringLangCode, SIZEOF(StringLangCode), lpNumberOfBytes, NIL);

        SetFilePointer(AIF, StringOffset, NIL, FILE_BEGIN);
        ReadFile(AIF, CapLength, SIZEOF(CapLength), lpNumberOfBytes, NIL);
        CapLength := CapLength div 4;
        GetMem(Cap, CapLength + 1);
        Cap[CapLength] := #0;
        ReadFile(AIF, Cap^, CapLength, lpNumberOfBytes, NIL);

        Item := AddItemStrings;
        Strings[Item] := Cap;
        StringsLangCode[Item] := StringLangCode;

        FreeMem(Cap, CapLength + 1);
      end;

      IconTrailer := SetFilePointer(
          AIF,
          TrailerOffset + SIZEOF(StringCount) + (SIZEOF(StringOffset) +
            SIZEOF(StringLangCode))*StringCount,
          NIL,
          FILE_BEGIN) + SIZEOF(IconCount);
      ReadFile(AIF, IconCount, SIZEOF(IconCount), lpNumberOfBytes, NIL);
      IconCount := IconCount div 2;

      if SetFilePointer(AIF, 0, NIL, FILE_CURRENT) < GetFileSize(AIF, NIL) then
      begin
        SetFilePointer(
          AIF,
          IconTrailer + IconCount*(SIZEOF(BitMapOffset) + SIZEOF(BitMapSize)),
          NIL,
          FILE_BEGIN);
        ReadFile(AIF, Constant, SIZEOF(Constant), lpNumberOfBytes, NIL);
        ReadFile(AIF, Embeddable, SIZEOF(Embeddable), lpNumberOfBytes, NIL);
        ReadFile(AIF, NewFile, SIZEOF(NewFile), lpNumberOfBytes, NIL);
        ReadFile(AIF, Hidden, SIZEOF(Hidden), lpNumberOfBytes, NIL);

        case Constant of
          $00000003:
          begin
            UseReserved := ReadFile(AIF, Reserved, SIZEOF(Reserved), lpNumberOfBytes, NIL);
          end;
        end;                                              

        MIMETrailer := SetFilePointer(AIF, 0, NIL, FILE_CURRENT);

        // MIME
        if MIMETrailer < GetFileSize(AIF, NIL) then
        begin
          ReadFile(AIF, MIMEID, SIZEOF(MIMEID), lpNumberOfBytes, NIL);
          ReadFile(AIF, MIMECount, SIZEOF(MIMECount), lpNumberOfBytes, NIL);
          MIMECount := MIMECount div 2;
          if MIMECount > 0 then
          begin
            for i:= 1 to MIMECount do
            begin
              SetFilePointer(
                AIF,
                MIMETrailer + SIZEOF(MIMEID) + SIZEOF(MIMECount) +
                  (SIZEOF(MIMEOffset) + SIZEOF(MIMELangCode))*(DWORD(i)-1),
                NIL,
                FILE_BEGIN);
              ReadFile(AIF, MIMEOffset, SIZEOF(MIMEOffset), lpNumberOfBytes, NIL);
              ReadFile(AIF, MIMELangCode, SIZEOF(MIMELangCode), lpNumberOfBytes, NIL);

              SetFilePointer(AIF, MIMEOffset, NIL, FILE_BEGIN);
              ReadFile(AIF, CapLength, SIZEOF(CapLength), lpNumberOfBytes, NIL);
              CapLength := (CapLength - 2) div 4;
              GetMem(Cap, CapLength + 1);
              ReadFile(AIF, Cap^, CapLength, lpNumberOfBytes, NIL);
              Cap[CapLength] := #0;

              Item := AddItemMimeType;
              MimeType[Item] := Cap;   
              MimeTypeLangCode[Item] := MIMELangCode;

              FreeMem(Cap, CapLength + 1);
            end;

            SetFilePointer(
              AIF,
              MIMETrailer + SIZEOF(MIMEID) + SIZEOF(MIMECount) +
                (SIZEOF(MIMEOffset) + SIZEOF(MIMELangCode))*MIMECount,
              NIL,
              FILE_BEGIN);
            ReadFile(AIF, MIMEEnd, SIZEOF(MIMEEnd), lpNumberOfBytes, NIL);
          end;
        end;
      end;

      // extract mbm
      if FMBMDataSize <> 0 then
        FreeMem(FMBMData, FMBMDataSize);
      FMBMDataSize := 0;

      // calc trailer offset
      TrailerOffset := 20;
      i := 0;
      while (i < IconCount * 2) do
      begin
        SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
        ReadFile(AIF, BufferSize, SIZEOF(BufferSize), lpNumberOfBytes, NIL);
        TrailerOffset := TrailerOffset + BufferSize;
        inc(i);
      end;

      if MBMToData then
      begin
        BufferSize := TrailerOffset - 20;
        MBMDataPosition := SIZEOF(MBMHeader);
        FMBMDataSize := SIZEOF(MBMHeader) + BufferSize;
        GetMem(FMBMData, SIZEOF(MBMHeader) + BufferSize);
        MoveMemory(FMBMData, @MBMHeader, SIZEOF(MBMHeader));
                                   
        // copy mbm data
        SetFilePointer(AIF, 20, NIL, FILE_BEGIN);
        while (BufferSize > 0) do
        begin
          ReadFile(AIF, Buffer, SIZEOF(Buffer), lpNumberOfBytes, NIL);
          if BufferSize >= lpNumberOfBytes then
          begin
            MoveMemory(Pointer(Integer(FMBMData) + MBMDataPosition), @Buffer, lpNumberOfBytes);
            MBMDataPosition := MBMDataPosition + Integer(lpNumberOfBytes);
            BufferSize := BufferSize - lpNumberOfBytes;
          end
          else
          begin
            MoveMemory(Pointer(Integer(FMBMData) + MBMDataPosition), @Buffer, BufferSize);
            MBMDataPosition := MBMDataPosition + Integer(BufferSize);
            Break;
          end;
        end;
                              
        // trailer
        IconCountInMBM := IconCount * 2;
        FMBMDataSize := FMBMDataSize + SIZEOF(IconCountInMBM) +
          Integer(IconCountInMBM)*(SIZEOF(BitMapOffset) + SIZEOF(BitMapOffset)); 
        ReAllocMem(FMBMData, FMBMDataSize);
        MoveMemory(Pointer(Integer(FMBMData) + MBMDataPosition), @IconCountInMBM, SIZEOF(IconCountInMBM));
        MBMDataPosition := MBMDataPosition + SIZEOF(IconCountInMBM);
        for i:= 1 to IconCountInMBM do
        begin
          SetFilePointer(
            AIF,
            IconTrailer + (SIZEOF(BitMapOffset) + SIZEOF(BitMapSize))*(DWORD(i)-1),
            NIL,
            FILE_BEGIN);
          // 1x
          ReadFile(AIF, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
          ReadFile(AIF, BitMapSize, SIZEOF(BitMapSize), lpNumberOfBytes, NIL);
          MoveMemory(Pointer(Integer(FMBMData) + MBMDataPosition), @BitMapOffset, SIZEOF(BitMapOffset));
          MBMDataPosition := MBMDataPosition + SIZEOF(BitMapOffset);
          // 2x
          SetFilePointer(AIF, BitMapOffset, NIL, FILE_BEGIN);
          ReadFile(AIF, BufferSize, SIZEOF(BufferSize), lpNumberOfBytes, NIL);
          BitMapOffset := BitMapOffset + BufferSize;
          MoveMemory(Pointer(Integer(FMBMData) + MBMDataPosition), @BitMapOffset, SIZEOF(BitMapOffset));
          MBMDataPosition := MBMDataPosition + SIZEOF(BitMapOffset);
        end;  
      end else // MBMToData
      begin
        MBM := OpenFile(PCHAR(FMBMFileName), lpReOpenBuff,
          OF_CREATE or OF_READWRITE);

        if MBM <> HFILE_ERROR then
        begin
          WriteFile(MBM, MBMHeader, SIZEOF(MBMHeader), lpNumberOfBytes, NIL);
          WriteFile(MBM, TrailerOffset, SIZEOF(TrailerOffset), lpNumberOfBytes, NIL);

          // copy mbm data
          SetFilePointer(AIF, 20, NIL, FILE_BEGIN);
          BufferSize := TrailerOffset - 20;
          while (BufferSize > 0) do
          begin
            ReadFile(AIF, Buffer, SIZEOF(Buffer), lpNumberOfBytes, NIL);
            if BufferSize >= lpNumberOfBytes then
            begin
              WriteFile(MBM, Buffer, lpNumberOfBytes, lpNumberOfBytes, NIL);
              BufferSize := BufferSize - lpNumberOfBytes;
            end
            else
            begin
              WriteFile(MBM, Buffer, BufferSize, lpNumberOfBytes, NIL);
              Break;
            end;
          end;

          // trailer
          SetFilePointer(MBM, TrailerOffset, NIL, FILE_BEGIN);
          IconCountInMBM := IconCount * 2;
          WriteFile(MBM, IconCountInMBM, SIZEOF(IconCountInMBM), lpNumberOfBytes, NIL);
          for i:= 1 to IconCountInMBM do
          begin
            SetFilePointer(
              AIF,
              IconTrailer + (SIZEOF(BitMapOffset) + SIZEOF(BitMapSize))*(DWORD(i)-1),
              NIL,
              FILE_BEGIN);
            // 1x
            ReadFile(AIF, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
            ReadFile(AIF, BitMapSize, SIZEOF(BitMapSize), lpNumberOfBytes, NIL);
            WriteFile(MBM, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
            // 2x
            SetFilePointer(AIF, BitMapOffset, NIL, FILE_BEGIN);
            ReadFile(AIF, BufferSize, SIZEOF(BufferSize), lpNumberOfBytes, NIL);
            BitMapOffset := BitMapOffset + BufferSize;
            WriteFile(MBM, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
          end;

          CloseHandle(MBM);
        end;
      end; // else MBMToData
    end; // AIF_FORMAT_EPOC

else
  begin
    CloseHandle(AIF);
    Exit;  // not supported this format
  end;

end; // case UIDS[1]

  CloseHandle(AIF);
  Result := True;
end;

function TAifIcon.SaveToFile;
var
  AIF: HFILE;
  MBM: HFILE;
  lpReOpenBuff: _OFSTRUCT;
  lpNumberOfBytes: DWORD;
  i: Integer;
  Buffer: array[0..1023] of BYTE;
  BufferSize: DWORD;
  TrailerOffset: DWORD;
  StringTrailer: DWORD;
  MIMETrailer: DWORD;
  StringCount: BYTE;
  StringLangCode: WORD;
  CapLength: BYTE;
  IconCountInMBM: DWORD;
  BitMapOffset: DWORD;
  BitMapSize: DWORD;
  BitMapWidth: WORD;
  MIMECount: BYTE;
  MIMELangCode: WORD;
  TmpName: String;
  lpTempPath: array[0..MAX_PATH-1] of CHAR;
begin
  Result := False;

  if UseMBMData and (FMBMDataSize > 0) then
  begin
    GetTempPath(MAX_PATH, lpTempPath);
    TmpName := String(lpTempPath) + 'mbm$$';

    MBM := OpenFile(PCHAR(TmpName), lpReOpenBuff, OF_CREATE or OF_WRITE);
    if MBM = HFILE_ERROR then
      Exit;
    WriteFile(MBM, FMBMData^, FMBMDataSize, lpNumberOfBytes, NIL);
    CloseHandle(MBM);
  end else
    TmpName := FMBMFileName;

    
  MBM := OpenFile(PCHAR(TmpName), lpReOpenBuff, OF_READ);

  if GetFileSize(MBM, NIL) <= 0 then
  begin
    CloseHandle(MBM);
    Exit;
  end;

  AIF := OpenFile(PCHAR(FileName), lpReOpenBuff, OF_CREATE or OF_READWRITE);
  if AIF = HFILE_ERROR then
  begin
    CloseHandle(MBM);
    Exit;
  end;
  SetEndOfFile(AIF);

  UIDS[4] := CalcUID4(UIDS[1], UIDS[2], UIDS[3]);

  // calc Trailer Offset
  SetFilePointer(MBM, 16, NIL, FILE_BEGIN);
  ReadFile(MBM, TrailerOffset, SIZEOF(TrailerOffset), lpNumberOfBytes, NIL);
  SetFilePointer(MBM, TrailerOffset, NIL, FILE_BEGIN);
  ReadFile(MBM, IconCountInMBM, SIZEOF(IconCountInMBM), lpNumberOfBytes, NIL);
  ReadFile(MBM, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
  TrailerOffset := 20;
  i := 0;
  while (DWORD(i) < IconCountInMBM) do
  begin
    SetFilePointer(MBM, BitMapOffset, NIL, FILE_BEGIN);
    ReadFile(MBM, BitmapSize, SIZEOF(BitmapSize), lpNumberOfBytes, NIL);
    BitMapOffset := BitMapOffset + BitmapSize;
    TrailerOffset := TrailerOffset + BitmapSize;
    inc(i);
  end;
  StringTrailer := TrailerOffset;
  for i:= 0 to CountItemStrings - 1 do
    TrailerOffset := TrailerOffset + SIZEOF(CapLength) + DWORD(Length(Strings[i]));
  MIMETrailer := TrailerOffset;
  for i:= 0 to CountItemMimeType - 1 do
    TrailerOffset := TrailerOffset + SIZEOF(CapLength) + DWORD(Length(MimeType[i]));

  // write info to AIF
  WriteFile(AIF, UIDS, SIZEOF(UIDS), lpNumberOfBytes, NIL);
  WriteFile(AIF, TrailerOffset, SIZEOF(TrailerOffset), lpNumberOfBytes, NIL);

  // copy data
  SetFilePointer(MBM, 20, NIL, FILE_BEGIN);
  BufferSize := StringTrailer - 20;
  while (BufferSize > 0) do
  begin
    ReadFile(MBM, Buffer, SIZEOF(Buffer), lpNumberOfBytes, NIL);
    if BufferSize >= lpNumberOfBytes then
    begin
      WriteFile(AIF, Buffer, lpNumberOfBytes, lpNumberOfBytes, NIL);
      BufferSize := BufferSize - lpNumberOfBytes;
    end
    else
    begin
      WriteFile(AIF, Buffer, BufferSize, lpNumberOfBytes, NIL);
      Break;
    end;
  end;

  // write strings
  StringCount := CountItemStrings * 2;
  SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
  WriteFile(AIF, StringCount, SIZEOF(StringCount), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(StringCount);
  for i:= 0 to CountItemStrings - 1 do
  begin
    CapLength := Length(Strings[i]);
    StringLangCode := StringsLangCode[i];
    SetFilePointer(AIF, StringTrailer, NIL, FILE_BEGIN);
    CapLength := CapLength * 4;
    WriteFile(AIF, CapLength, SIZEOF(CapLength), lpNumberOfBytes, NIL);
    WriteFile(AIF, PCHAR(Strings[i])^, Length(Strings[i]), lpNumberOfBytes, NIL);
    
    SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
    WriteFile(AIF, StringTrailer, SIZEOF(StringTrailer), lpNumberOfBytes, NIL);
    WriteFile(AIF, StringLangCode, SIZEOF(StringLangCode), lpNumberOfBytes, NIL);
    TrailerOffset := TrailerOffset + SIZEOF(StringTrailer) + SIZEOF(StringLangCode);
    StringTrailer := StringTrailer + SIZEOF(CapLength) + DWORD(Length(Strings[i]));
  end;

  // write offset of Icon's
  WriteFile(AIF, IconCountInMBM, SIZEOF(BYTE), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(BYTE);
  i := 0;
  BitMapOffset := 20;
  while (DWORD(i) < IconCountInMBM div 2) do
  begin
    // 1x
    SetFilePointer(MBM, BitMapOffset, NIL, FILE_BEGIN);
    ReadFile(MBM, BitmapSize, SIZEOF(BitmapSize), lpNumberOfBytes, NIL);
    // write
    SetFilePointer(MBM, 4, NIL, FILE_CURRENT);
    ReadFile(MBM, BitMapWidth, SIZEOF(BitMapWidth), lpNumberOfBytes, NIL);
    WriteFile(AIF, BitMapOffset, SIZEOF(BitMapOffset), lpNumberOfBytes, NIL);
    WriteFile(AIF, BitMapWidth, SIZEOF(BitMapWidth), lpNumberOfBytes, NIL);
    TrailerOffset := TrailerOffset + SIZEOF(BitMapOffset) + SIZEOF(BitMapWidth);
    BitMapOffset := BitMapOffset + BitmapSize;
    // 2x
    SetFilePointer(MBM, BitMapOffset, NIL, FILE_BEGIN);
    ReadFile(MBM, BitmapSize, SIZEOF(BitmapSize), lpNumberOfBytes, NIL);
    BitMapOffset := BitMapOffset + BitmapSize;
    inc(i);
  end;

  // write Data
  WriteFile(AIF, Constant, SIZEOF(Constant), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(Constant);
  WriteFile(AIF, Embeddable, SIZEOF(Embeddable), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(Embeddable);
  WriteFile(AIF, NewFile, SIZEOF(NewFile), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(NewFile);
  WriteFile(AIF, Hidden, SIZEOF(Hidden), lpNumberOfBytes, NIL);
  TrailerOffset := TrailerOffset + SIZEOF(Hidden);

  if UseReserved then
  begin
    WriteFile(AIF, Reserved, SIZEOF(Reserved), lpNumberOfBytes, NIL);
    TrailerOffset := TrailerOffset + SIZEOF(Reserved);
  end;
      
  // write MIME
  MIMECount := CountItemMimeType * 2;
  if MIMECount > 0 then
  begin
    SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
    WriteFile(AIF, MIMEID, SIZEOF(MIMEID), lpNumberOfBytes, NIL);
    TrailerOffset := TrailerOffset + SIZEOF(MIMEID);
    WriteFile(AIF, MIMECount, SIZEOF(MIMECount), lpNumberOfBytes, NIL);
    TrailerOffset := TrailerOffset + SIZEOF(MIMECount);
    for i:= 0 to CountItemMimeType - 1 do
    begin
      MIMELangCode := MimeTypeLangCode[i];
      CapLength := Length(MimeType[i]) * 4 + 2;
      
      SetFilePointer(AIF, MIMETrailer, NIL, FILE_BEGIN);
      WriteFile(AIF, CapLength, SIZEOF(CapLength), lpNumberOfBytes, NIL);
      WriteFile(AIF, PCHAR(MimeType[i])^, Length(MimeType[i]), lpNumberOfBytes, NIL);

      SetFilePointer(AIF, TrailerOffset, NIL, FILE_BEGIN);
      WriteFile(AIF, MIMETrailer, SIZEOF(MIMETrailer), lpNumberOfBytes, NIL);
      WriteFile(AIF, MIMELangCode, SIZEOF(MIMELangCode), lpNumberOfBytes, NIL);

      MIMETrailer := MIMETrailer + SIZEOF(CapLength) + DWORD(Length(MimeType[i]));
      TrailerOffset := TrailerOffset + SIZEOF(MIMETrailer) + SIZEOF(MIMELangCode);
    end;
    WriteFile(AIF, MIMEEnd, SIZEOF(MIMEEnd), lpNumberOfBytes, NIL);
  end;

  CloseHandle(MBM);
  CloseHandle(AIF);

  if UseMBMData then
    DeleteFile(PCHAR(TmpName));

  Result := True;
end; 

end.
