unit SX1DiskA;

interface
  uses Windows, ApiFunc;


type
  THeaderROLF = packed record
    ROLF: array[0..35] of CHAR;
    dwSize: DWORD;
    dwReserved: array[0..87] of CHAR;
  end;
  PHeaderROLF = ^THeaderROLF;

  function InitHeaderROLF(lpROLF: Pointer): LONGBOOL;



type
  THeaderBeginROLF = packed record
    dwReserved1: array[0..7] of CHAR;
    lpSectDirectories: DWORD;
    szSectDirectories: DWORD;
    lpSectFiles: DWORD;
    szSectFiles: DWORD;
    dwReserved2: array[0..11] of CHAR;
    dwSize: DWORd;
    dwReserved3: array[0..3] of CHAR;
  end;
  PHeaderBeginROLF = ^THeaderBeginROLF;
  
  function InitHeaderBeginRolf(lpBeginROLF: Pointer): LONGBOOL;

  

const
  ID_SECTION = $0CFF;
  ID_DATA = $010E;
  ID_OBJECT = $100E;

  EXT_DIRECTORIES = $00000000;
  EXT_FILES = $00000010;

type
  THeaderSection = packed record
    dwSize: WORD;
    dwReserved: WORD; // ID_SECTION
    dwFlag1: DWORD;
    dwFlag2: DWORD;
  end;

  THeaderObject = packed record
    dwSize: WORD;
    dwReserved: WORD; // ID_OBJECT
    Size: DWORD;
    Offset: DWORD;
    dwFlag: BYTE; // $FF
    LengthName: BYTE;
    Name: WideString; 
  end;

  HROLF = UINT;
  
  TROLFDATA = packed record
    hFile: HFILE;
    uFlag: DWORD;
  end;
  PROLFDATA = ^TROLFDATA;

const
  ROLF_ERROR = HROLF(0);

  ROLF_PACKED = $00000010; 
  ROLF_UNPACKED = $00001000;

  
  function CreateROLF(lpFileName: String; uFlag: DWORD): HROLF;
  function CheckROLF(ROLF: HROLF): LONGBOOL;
  function Checksize(ROLF: HROLF; MaxSize: DWORD): LONGBOOL;
  function UnPackedROLF(ROLF: HROLF; lpDirectory: String; uFlags: DWORD; StartOffset: DWORD): LONGBOOL;
  function PackedROLF(ROLF: HROLF; lpDirectory: String): LONGBOOL;
  function FreeROLF(ROLF: HROLF): LONGBOOL;

implementation

function InitHeaderROLF;
var
  pROLF: PHeaderROLF;
begin
  Result := False;
  pROLF := PHeaderROLF(lpROLF);
  if (pROLF <> NIL) then
  with pROLF^ do
  begin
    ROLF := #$52#$4F#$4C#$46#$5F#$69#$6D#$61#$67#$65;
    dwSize := 0;               
    dwReserved := #$00#$00#$0A#$04;
    Result := True;
  end;
end;

function InitHeaderBeginRolf;
var
  pBeginROLF: PHeaderBeginROLF;
begin
  Result := False;
  pBeginROLF := PHeaderBeginROLF(lpBeginROLF);
  if (pBeginROLF <> NIL) then
  with pBeginROLF^ do
  begin
    dwReserved1 := #$52#$4F#$4C#$46#$2C#$FF#$00#$01;
    lpSectDirectories := 0;
    szSectDirectories := 0;
    lpSectFiles := 0;
    szSectFiles := 0;
    ZeroMemory(@dwReserved2, SIZEOF(dwReserved2));
    dwSize := 0;
    ZeroMemory(@dwReserved3, SIZEOF(dwReserved3));
    Result := True;
  end;
end;

function CreateROLF;
var
  RolfData: PROLFDATA;
  lpReOpenBuff: _OFSTRUCT;
  HF: HFILE;
begin
  Result := ROLF_ERROR;
  case uFlag of
    ROLF_UNPACKED:HF := OpenFile(PCHAR(lpFileName), lpReOpenBuff, OF_READ);
    ROLF_PACKED:HF := OpenFile(PCHAR(lpFileName), lpReOpenBuff, OF_CREATE or OF_READWRITE);
    else HF := HFILE_ERROR;
  end;
  if (HF <> HFILE_ERROR) then
  begin
    GetMem(RolfData, SIZEOF(TROLFDATA));
    RolfData^.hFile := HF;
    RolfData^.uFlag := uFlag;
    Result := HROLF(RolfData);
  end;
end;

function CheckROLF;
begin
  Result := (ROLF <> ROLF_ERROR) and (PROLFDATA(ROLF) <> NIL);
end;

function Checksize;
var
  lpFileName: LPCSTR;
begin
  Result := False;
  if (MaxSize>0) and (CheckROLF(ROLF)) then
  with PROLFDATA(ROLF)^ do
  if (uFlag = ROLF_PACKED) and (GetFileSize(hFile, NIL) > MaxSize) then
  begin
    GetMem(lpFileName, MAX_PATH);
    GetModuleFileName(hFile, lpFileName, MAX_PATH);
    FreeROLF(ROLF);
    DeleteFile(lpFileName);
    FreeMem(lpFileName);
    Result := True;
  end;  
end;

function FreeROLF;
begin
  Result := False;
  if CheckROLF(ROLF) then
  begin
    CloseHandle(PROLFDATA(ROLF)^.hFile);
    FreeMem(PROLFDATA(ROLF));
    Result := True;
  end;
end;

procedure ReadSections(hFile: HFILE; Position: DWORD; ParentDirectory: String; uFlags: DWORD);
var
  HeaderSection: THeaderSection;
  HeaderObject: THeaderObject;
  lpNumberOfBytes: DWORD;
  dwSize, dwOffset: DWORD;
  ReOffset: DWORD;
  ParentDir: String;
  FOut: UINT;
  lpReOpenBuff: _OFSTRUCT;
  lpBuffer: Pointer;
begin
  SetFilePointer(hFile, Position, NIL, FILE_BEGIN); 
  if not ReadFile(hFile, HeaderSection, SIZEOF(THeaderSection), lpNumberOfBytes, NIL) then exit;
  if (HeaderSection.dwReserved <> ID_SECTION) then exit;

  if (HeaderSection.dwFlag2 > 0) and (uFlags and EXT_FILES = EXT_FILES) then
  begin
    ReOffset := SetFilePointer(hFile, 0, NIL, FILE_CURRENT);
    //  
    SetFilePointer(hFile, HeaderSection.dwFlag1 + SIZEOF(THeaderROLF), NIL, FILE_BEGIN);
    dwSize := 0;
    while (dwSize < HeaderSection.dwFlag2) do
    begin
      if not ReadFile(hFile, HeaderObject, 14, lpNumberOfBytes, NIL) then exit;
      if (HeaderObject.dwReserved <> ID_DATA) then exit;
      if (HeaderObject.LengthName = 0) then exit;
      SetLength(HeaderObject.Name, HeaderObject.LengthName);
      if not ReadFile(hFile, Pointer(HeaderObject.Name)^, HeaderObject.LengthName*2, lpNumberOfBytes, NIL) then exit;
      dwOffset := SetFilePointer(hFile, 0, NIL, FILE_CURRENT);
      dwSize := dwSize + HeaderObject.dwSize;
      if (dwSize mod 4 <> 0) then
      begin
        dwSize := dwSize + 2;
        dwOffset := dwOffset + 2;
      end;
      SetFilePointer(hFile, SIZEOF(THeaderROLF) + HeaderObject.Offset, NIL, FILE_BEGIN);
      FOut := OpenFile(PCHAR(ParentDirectory+WideCharToString(PWCHAR(HeaderObject.Name))),
        lpReOpenBuff, OF_CREATE or OF_READWRITE);
      if (FOut <> HFILE_ERROR) then
      begin
        GetMem(lpBuffer, HeaderObject.Size);
        ReadFile(hFile, lpBuffer^, HeaderObject.Size, lpNumberOfBytes, NIL);
        WriteFile(FOut, lpBuffer^, lpNumberOfBytes, lpNumberOfBytes, NIL);
        FreeMem(lpBuffer);
        CloseHandle(FOut);
      end;
      SetFilePointer(hFile, dwOffset, NIL, FILE_BEGIN);
    end;
    SetFilePointer(hFile, ReOffset, NIL, FILE_BEGIN);
  end;

  //   
  dwSize := SIZEOF(THeaderSection);
  while (dwSize < HeaderSection.dwSize) do
  begin
    if not ReadFile(hFile, HeaderObject, 14, lpNumberOfBytes, NIL) then exit;
    if (HeaderObject.dwReserved <> ID_OBJECT) then exit;
    if (HeaderObject.LengthName = 0) then exit;
    SetLength(HeaderObject.Name, HeaderObject.LengthName);
    if not ReadFile(hFile, Pointer(HeaderObject.Name)^, HeaderObject.LengthName*2, lpNumberOfBytes, NIL) then exit; 
    dwOffset := SetFilePointer(hFile, 0, NIL, FILE_CURRENT);
    dwSize := dwSize + HeaderObject.dwSize;
    if (dwSize mod 4 <> 0) then
    begin
      dwSize := dwSize + 2;
      dwOffset := dwOffset + 2;
    end;
    ParentDir := ParentDirectory + WideCharToString(PWCHAR(HeaderObject.Name)) + '\';
    CreateDirectory(PCHAR(ParentDir), NIL);
    if (HeaderObject.Offset>0) and (HeaderObject.Size >= SIZEOF(THeaderSection)) and
      (uFlags and EXT_DIRECTORIES = EXT_DIRECTORIES) then
        ReadSections(hFile, SIZEOF(THeaderROLF) + HeaderObject.Offset, ParentDir, uFlags);
    SetFilePointer(hFile, dwOffset, NIL, FILE_BEGIN);
  end;
end;

function UnPackedROLF;
var
  HeaderROLF: THeaderROLF;
  HeaderBeginROLF: THeaderBeginROLF;
  lpNumberOfBytesRead: DWORD;
begin
  Result := False;
  if (CheckROLF(ROLF)) then
  with PROLFDATA(ROLF)^ do
  if (uFlag = ROLF_UNPACKED) then
  begin
    SetFilePointer(hFile, StartOffset, NIL, FILE_BEGIN);
    if not ReadFile(hFile, HeaderRolf, SIZEOF(THeaderROLF), lpNumberOfBytesRead, NIL) then exit;
    if not ReadFile(hFile, HeaderBeginROLF, SIZEOF(THeaderBeginROLF), lpNumberOfBytesRead, NIL) then exit;
    CreateDirectory(PCHAR(lpDirectory), NIL);
    ReadSections(hFile,
      HeaderBeginROLF.lpSectDirectories + SIZEOF(THeaderROLF),
      lpDirectory,
      uFlags);
    Result := True;
  end;  
end;

procedure WriteSections(hFile: HFILE; Directory: String; lpSubDirectory: DWORD;
  var lpHeaderFiles, lpDataFiles: DWORD);
var
  HeaderSection: THeaderSection;
  HeaderObject: THeaderObject;
  lpFindFileData: _WIN32_FIND_DATA;
  hFindFile: DWORD;
  dwOffset: DWORD;
  lpSubDir: DWORD;
  lpNumberOfBytes: DWORD;
  SEEKFLAG, LASTFLAG: WORD;
  FILESEEKSIZE: DWORD;
  inp: DWORD;
  lpReOpenBuff: _OFSTRUCT;
  lpBuffer: Pointer;
  lpOffset: DWORD;
begin
  //  
  SEEKFLAG := $FFFF; LASTFLAG := 0;
  dwOffset := SetFilePointer(hFile, 0, NIL, FILE_CURRENT);

  //     
  SetFilePointer(hFile, SIZEOF(THeaderSection), NIL, FILE_CURRENT);

  //       
  hFindFile := FindFirstFile(PCHAR(Directory + '*.*'), lpFindFileData);
  if (hFindFile <> INVALID_HANDLE_VALUE) then
  repeat
    if (lpFindFileData.cFileName[0]<>'.') and
      (lpFindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY =
        FILE_ATTRIBUTE_DIRECTORY) then
    begin
      ZeroMemory(@HeaderObject, SIZEOF(THeaderObject));
      HeaderObject.dwReserved := ID_OBJECT;
      HeaderObject.dwFlag := $FF;
      HeaderObject.LengthName := lstrlenA(lpFindFileData.cFileName);
      HeaderObject.dwSize := 14 + HeaderObject.LengthName*2;
      SetLength(HeaderObject.Name, HeaderObject.LengthName);
      StringToWideChar(lpFindFileData.cFileName, LPWSTR(HeaderObject.Name), HeaderObject.LengthName*2);
      WriteFile(hFile, HeaderObject, 14, lpNumberOfBytes, NIL);
      WriteFile(hFile, LPWSTR(HeaderObject.Name)^, HeaderObject.LengthName*2, lpNumberOfBytes, NIL);

      if ((SetFilePointer(hFile, 0, NIL, FILE_CURRENT) - dwOffset) mod 4 <> 0) then
      begin
        LASTFLAG := 2;
        WriteFile(hFile, SEEKFLAG, 2, lpNumberOfBytes, NIL);
      end else LASTFLAG := 0;
    end;
  until not FindNextFile(hFindFile, lpFindFileData);
  FindClose(hFindFile);

  //   
  HeaderSection.dwSize := (SetFilePointer(hFile, 0, NIL, FILE_CURRENT) - dwOffset) - LASTFLAG;

  //    
  SetFilePointer(hFile, (lpHeaderFiles + SIZEOF(THeaderROLF)), NIL, FILE_BEGIN);
  HeaderSection.dwFlag1 := lpHeaderFiles;
  HeaderSection.dwFlag2 := 0;
  hFindFile := FindFirstFile(PCHAR(Directory + '*.*'), lpFindFileData);
  if (hFindFile <> INVALID_HANDLE_VALUE) then
  repeat
    if (lpFindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0) then
    begin
      ZeroMemory(@HeaderObject, SIZEOF(THeaderObject));
      HeaderObject.dwReserved := ID_DATA;
      //    dump
      inp := OpenFile(LPCSTR(Directory + lpFindFileData.cFileName), lpReOpenBuff, OF_READ);
      if (inp <> HFILE_ERROR) then
      begin
        HeaderObject.Size := GetFileSize(inp, NIL);
        HeaderObject.Offset := lpDataFiles;
        lpOffset := SetFilePointer(hFile, 0, NIL, FILE_CURRENT);
        SetFilePointer(hFile, (lpDataFiles+SIZEOF(THeaderROLF)), NIL, FILE_BEGIN);
        GetMem(lpBuffer, HeaderObject.Size);
        SetFilePointer(inp, 0, NIL, FILE_BEGIN);
        ReadFile(inp, lpBuffer^, HeaderObject.Size, lpNumberOfBytes, NIL);
        WriteFile(hFile, lpBuffer^, lpNumberOfBytes, lpNumberOfBytes, NIL);
        FreeMem(lpBuffer);
        CloseHandle(inp);
        if (HeaderObject.Size mod 4 <> 0) then
        begin
          FILESEEKSIZE := HeaderObject.Size;
          while (FILESEEKSIZE mod 4 <> 0) do
          begin
            FILESEEKSIZE := FILESEEKSIZE + 1;
            WriteFile(hFile, BYTE(SEEKFLAG), 1, lpNumberOfBytes, NIL);
            lpDataFiles := lpDataFiles + 1;
          end;
        end;
        lpDataFiles := lpDataFiles + HeaderObject.Size;
        SetFilePointer(hFile, lpOffset, NIL, FILE_BEGIN);
      end;
      //   
      HeaderObject.dwFlag := $FF;
      HeaderObject.LengthName := lstrlenA(lpFindFileData.cFileName);
      HeaderObject.dwSize := 14 + HeaderObject.LengthName*2;
      SetLength(HeaderObject.Name, HeaderObject.LengthName);
      StringToWideChar(lpFindFileData.cFileName, LPWSTR(HeaderObject.Name), HeaderObject.LengthName*2);
      WriteFile(hFile, HeaderObject, 14, lpNumberOfBytes, NIL);
      WriteFile(hFile, LPWSTR(HeaderObject.Name)^, HeaderObject.LengthName*2, lpNumberOfBytes, NIL);
      HeaderSection.dwFlag2 := HeaderSection.dwFlag2 + HeaderObject.dwSize;
      if (HeaderSection.dwFlag2 mod 4 <> 0) then
      begin
        HeaderSection.dwFlag2 := HeaderSection.dwFlag2 + 2;
        WriteFile(hFile, SEEKFLAG, SIZEOF(WORD), lpNumberOfBytes, NIL);
      end;
    end;
  until not FindNextFile(hFindFile, lpFindFileData);
  FindClose(hFindFile);
  lpHeaderFiles := lpHeaderFiles + HeaderSection.dwFlag2;
  if (HeaderSection.dwFlag2=0) then HeaderSection.dwFlag1 := 0;

  //   
  SetFilePointer(hFile, dwOffset, NIL, FILE_BEGIN);
  
  //  
  HeaderSection.dwReserved := ID_SECTION;
  WriteFile(hFile, HeaderSection, SIZEOF(THeaderSection), lpNumberOfBytes, NIL);

  //       
  if (lpSubDirectory>0) then
  begin
    SetFilePointer(hFile, lpSubDirectory, NIL, FILE_BEGIN);
    WriteFile(hFile, HeaderSection.dwSize, SIZEOF(WORD), lpNumberOfBytes, NIL);
    SetFilePointer(hFile, 2, NIL, FILE_CURRENT);
    lpSubDir := dwOffset - SIZEOF(THeaderROLF);
    WriteFile(hFile, lpSubDir, SIZEOF(DWORD), lpNumberOfBytes, NIL);
  end;

  //     
  lpSubDir := dwOffset + SIZEOF(THeaderSection);
  SetFilePointer(hFile, (dwOffset + HeaderSection.dwSize + LASTFLAG), NIL, FILE_BEGIN);
  hFindFile := FindFirstFile(PCHAR(Directory + '*.*'), lpFindFileData);
  if (hFindFile <> INVALID_HANDLE_VALUE) then
  repeat
    if (lpFindFileData.cFileName[0]<>'.') and
      (lpFindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY =
        FILE_ATTRIBUTE_DIRECTORY) then
    begin
      lpSubDir := lpSubDir + 4;
      WriteSections(hFile, Directory + lpFindFileData.cFileName + '\',
        lpSubDir, lpHeaderFiles, lpDataFiles);
      lpSubDir := lpSubDir + 10 + DWORD(lstrlenA(lpFindFileData.cFileName)*2);
      if ((lpSubDir - dwOffset) mod 4 <> 0) then lpSubDir := lpSubDir + 2;
    end;
  until not FindNextFile(hFindFile, lpFindFileData);
  FindClose(hFindFile);
end;

function CalculationSize(Directory: String; FileOrDir: Boolean): DWORD;
const
  cFileOrDir: array[Boolean] of DWORD = (FILE_ATTRIBUTE_DIRECTORY, 0);
var
  lpFindFileData: _WIN32_FIND_DATA;
  hFindFile: DWORD;
  dwSize: DWORD;
begin
  if (FileOrDir) then dwSize := 0
  else dwSize := SIZEOF(THeaderSection);
  
  hFindFile := FindFirstFile(PCHAR(Directory + '*.*'), lpFindFileData);
  if (hFindFile <> INVALID_HANDLE_VALUE) then
  repeat
    if (lpFindFileData.cFileName[0]<>'.') and
      (lpFindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY =
        cFileOrDir[FileOrDir]) then
    begin
      dwSize := dwSize + (14 + DWORD(lstrlenA(lpFindFileData.cFileName))*2);
      if (dwSize mod 4 <> 0) then dwSize := dwSize + 2;
    end;
  until not FindNextFile(hFindFile, lpFindFileData);
  FindClose(hFindFile);

  hFindFile := FindFirstFile(PCHAR(Directory + '*.*'), lpFindFileData);
  if (hFindFile <> INVALID_HANDLE_VALUE) then
  repeat
    if (lpFindFileData.cFileName[0]<>'.') and
      (lpFindFileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY =
        FILE_ATTRIBUTE_DIRECTORY) then
    begin
      dwSize := dwSize +
        CalculationSize(Directory + lpFindFileData.cFileName + '\', FileOrDir);
    end;
  until not FindNextFile(hFindFile, lpFindFileData);
  FindClose(hFindFile);

  Result := dwSize;
end;

function PackedROLF;
var
  HeaderROLF: THeaderROLF;
  HeaderBeginROLF: THeaderBeginROLF;
  lpNumberOfBytes: DWORD;
  lpHeaderFiles, lpDataFiles: DWORD;
begin
  Result := False;
  if (CheckROLF(ROLF)) then
  with PROLFDATA(ROLF)^ do
  if (uFlag = ROLF_PACKED) then
  begin
    InitHeaderROLF(@HeaderROLF);
    InitHeaderBeginRolf(@HeaderBeginROLF);
    SetFilePointer(hFile, SIZEOF(THeaderROLF) + SIZEOF(THeaderBeginROLF), NIL, FILE_BEGIN);
    
    HeaderBeginROLF.lpSectDirectories := SetFilePointer(hFile, 0, NIL, FILE_CURRENT) - SIZEOF(THeaderROLF);
    HeaderBeginROLF.szSectDirectories := CalculationSize(lpDirectory, False);
    HeaderBeginROLF.lpSectFiles := HeaderBeginROLF.lpSectDirectories + HeaderBeginROLF.szSectDirectories;
    HeaderBeginROLF.szSectFiles := CalculationSize(lpDirectory, True);

    lpHeaderFiles := HeaderBeginROLF.lpSectFiles;
    lpDataFiles := HeaderBeginROLF.lpSectFiles + HeaderBeginROLF.szSectFiles;
    
    WriteSections(hFile, lpDirectory, 0, lpHeaderFiles, lpDataFiles);

    HeaderROLF.dwSize := GetFileSize(hFile, NIL) - SIZEOF(THeaderROLF);
    HeaderBeginROLF.dwSize := HeaderROLF.dwSize;

    SetFilePointer(hFile, 0, NIL, FILE_BEGIN);
    WriteFile(hFile, HeaderROLF, SIZEOF(THeaderROLF), lpNumberOfBytes, NIL);
    WriteFile(hFile, HeaderBeginROLF, SIZEOF(THeaderBeginROLF), lpNumberOfBytes, NIL);

    Result := True;
  end;
end;

end.
