unit SIS;

interface
  uses Windows, ApiFunc, ZLib;

{ SIS FORMAT SPECIFICATION }

type
  HSIS = UINT;
  HMSIS = UINT;

  _SIS_FILE_HEADER = packed record
    UID1: DWORD;
    UID2: DWORD;
    UID3: DWORD;
    UID4: DWORD;
    Checksum: WORD;
    NumberOfLanguages: WORD;
    NumberOfFiles: WORD;
    NumberOfRequisites: WORD;
    InstallationLanguage: WORD;
    InstallationFiles: WORD;
    InstallationDrive: WORD;
    NumberOfCapabilities: WORD;
    InstallerVersion: DWORD;
    Options: WORD;
    sisType: WORD;
    MajorVersion: WORD;
    MinorVersion: WORD;
    Variant: DWORD;
    lpLanguages: DWORD;
    lpFiles: DWORD;
    lpRequisites: DWORD;
    lpCertificates: DWORD;
    lpComponentName: DWORD;
    lpSignature: DWORD;
    lpCapabilities: DWORD;
    spInstalled: DWORD;
    spMaximumInstalledSpace: DWORD;              
    Reserved: array[0..3] of DWORD;
  end;
  TSisFileHeader = _SIS_FILE_HEADER;
  PTSisFileHeader = ^TSisFileHeader;

  _SIS_EVENT = function(SisHandle: UINT; ID: UINT; lParam: LPARAM; wParam: WPARAM): UINT of object; stdcall;
  TSisEvent = _SIS_EVENT;
  PTSisEvent = ^TSisEvent;

const
  INVALID_SIS_HANDLE = HSIS(0);

{ ProcessType - ID on _SIS_EVENT}
  SIS_PRCO_TYPE = $000F0000;
  // not used lParam & wParam
  SIS_PROC_TYPE_FREE = SIS_PRCO_TYPE+1;
  // lParam - pointer on SIS name
  SIS_PROC_TYPE_BUILD = SIS_PRCO_TYPE+2;
  SIS_PROC_TYPE_EXTRACT = SIS_PRCO_TYPE+3;
  SIS_PROC_TYPE_FILE = SIS_PRCO_TYPE+4;

{ EPOC UIDs }
  EPOC_UID_DEF_APP = $10000000; { UID 1 }
  EPOC_UID_REL_3 = $1000006D; { UID 2 }
  EPOC_UID_REL_6 = $10003A12; { UID 2 }
  EPOC_UID_ALWAYS = $10000419; { UID 3 }

  EPOC_INS_DRIVE_REL_3 = $0000;
  EPOC_INS_DRIVE_REL_6 = $0021;

  EPOC_INS_VER_REL_3_1 = $00000044;
  EPOC_INS_VER_REL_3_2 = $00000064;
  EPOC_INS_VER_REL_6 = $000000C8;   

{ Flags }
  SIS_NODIR = 0;
  SIS_SAVEDIR = 1;
  
{ Options }
  SIS_OPT_IU = $0001; // IsUnicode (EPOC release 5)
  SIS_OPT_ID = $0002; // IsDistributable (EPOC release 5)
  SIS_OPT_NC = $0008; // NOCOMPRESS	NoCompress (EPOC release 6)
  SIS_OPT_SH = $0010; // SHUTDOWNAPPS	ShutdownApps (EPOC release 6)

{ Type }
  SIS_TYPE_SA = $0000; // SISAPP	Contains an application (the default)
  SIS_TYPE_SY = $0001; // SISSYSTEM	Contains a shared/system component such as a DLL or OPX (EPOC release 6)
  SIS_TYPE_SO = $0002; // SISOPTION	Contains an optional component, independently selectable by the user (EPOC release 6)
  SIS_TYPE_SC = $0003; // SISCONFIG	Configures an existing application or service (EPOC release 6)
  SIS_TYPE_SP = $0004; // SISPATCH	Patches an existing component (EPOC release 6)
  SIS_TYPE_SU = $0005; // SISUPGRADE	Upgrades an existing component (EPOC release 6)

type
  _LANGUAGE = packed record
    ID: WORD;
  end;
  TLanguage = _LANGUAGE;
  PTLanguage = ^TLanguage;

const
{ Languages }
  SIS_LANG_EN = $0001; // UK English
  SIS_LANG_FR = $0002; // French
  SIS_LANG_GE = $0003; // German
  SIS_LANG_SP = $0004; // Spanish
  SIS_LANG_IT = $0005; // Italian
  SIS_LANG_SW = $0006; // Swedish
  SIS_LANG_DA = $0007; // Danish
  SIS_LANG_NO = $0008; // Norwegian
  SIS_LANG_FI = $0009; // Finnish
  SIS_LANG_AM = $000A; // American English
  SIS_LANG_SWF = $000B; // Swiss French
  SIS_LANG_SG = $000C; // Swiss German
  SIS_LANG_PO = $000D; // Portuguese
  SIS_LANG_TU = $000E; // Turkish
  SIS_LANG_IC = $000F; // Icelandic
  SIS_LANG_RU = $0010; // Russian
  SIS_LANG_HU = $0011; // Hungarian
  SIS_LANG_DU = $0012; // Dutch
  SIS_LANG_BL = $0013; // Belgian Flemish
  SIS_LANG_AU = $0014; // Australian English
  SIS_LANG_BG = $0015; // Belgian French
  SIS_LANG_AS = $0016; // Austrian German
  SIS_LANG_NZ = $0017; // New Zealand English
  SIS_LANG_IN = $0018; // International French
  SIS_LANG_CS = $0019; // Czech
  SIS_LANG_SK = $001A; // Slovak
  SIS_LANG_PL = $001B; // Polish
  SIS_LANG_SL = $001C; // Slovenian
  SIS_LANG_TC = $001D; // Taiwan Chinese
  SIS_LANG_HK = $001E; // Hong Kong Chinese
  SIS_LANG_ZH = $001F; // PRC Chinese
  SIS_LANG_JA = $0020; // Japanese
  SIS_LANG_TH = $0021; // Thai
  SIS_LANG_AF = $0022; // Afrikaans
  SIS_LANG_SQ = $0023; // Albanian
  SIS_LANG_AH = $0024; // Amharic
  SIS_LANG_AR = $0025; // Arabic
  SIS_LANG_HY = $0026; // Armenian
  SIS_LANG_TL = $0027; // Tagalog
  SIS_LANG_BE = $0028; // Belarussian
  SIS_LANG_BN = $0029; // Bengali
  SIS_LANG_BGR = $002A; // Bulgarian
  SIS_LANG_MY = $002B; // Burmese
  SIS_LANG_CA = $002C; // Catalan
  SIS_LANG_HR = $002D; // Croatian
  SIS_LANG_CE = $002E; // Canadian English
  SIS_LANG_IE = $002F; // International English
  SIS_LANG_SF = $0030; // South African English
  SIS_LANG_ET = $0031; // Estonian
  SIS_LANG_FA = $0032; // Farsi
  SIS_LANG_CF = $0033; // Canadian French
  SIS_LANG_GD = $0034; // Scots Gaelic
  SIS_LANG_KA = $0035; // Georgian
  SIS_LANG_EL = $0036; // Greek
  SIS_LANG_CG = $0037; // Cyprus Greek
  SIS_LANG_GU = $0038; //  Gujarati
  SIS_LANG_HE = $0039; // Hebrew
  SIS_LANG_HI = $003A; // Hindi
  SIS_LANG_IND = $003B; // Indonesian
  SIS_LANG_GA = $003C; // Irish
  SIS_LANG_SZ = $003D; // Swiss Italian
  SIS_LANG_KN = $003E; // Kannada
  SIS_LANG_KK = $003F; // Kazakh
  SIS_LANG_KM = $0040; // Khmer
  SIS_LANG_KO = $0041; // Korean
  SIS_LANG_LO = $0042; // Laothian
  SIS_LANG_LV = $0043; // Latvian
  SIS_LANG_LT = $0044; // Lithuanian
  SIS_LANG_MK = $0045; // Macedonian
  SIS_LANG_MS = $0046; // Malay
  SIS_LANG_ML = $0047; // Malayalam
  SIS_LANG_MR = $0048; // Marathi
  SIS_LANG_MO = $0049; // Moldavian
  SIS_LANG_MN = $004A; // Mongolian
  SIS_LANG_NN = $004B; // Norwegian-Nynorsk
  SIS_LANG_BP = $004C; // Brazilian Portuguese
  SIS_LANG_PA = $004D; // Punjabi
  SIS_LANG_RO = $004E; // Romanian
  SIS_LANG_SR = $004F; // Serbian
  SIS_LANG_SI = $0050; // Sinhalese
  SIS_LANG_SO = $0051; // Somali
  SIS_LANG_OS = $0052; // International Spanish
  SIS_LANG_LS = $0053; // Latin American Spanish
  SIS_LANG_SH = $0054; // Swahili
  SIS_LANG_FS = $0055; // Finland Swedish
  { not #86 lang }
  SIS_LANG_TA = $0057; // Tamil
  SIS_LANG_TE = $0058; // Telugu
  SIS_LANG_BO = $0059; // Tibetan
  SIS_LANG_TI = $005A; // Tigrinya
  SIS_LANG_CT = $005B; // Cyprus Turkish
  SIS_LANG_TK = $005C; // Turkmen
  SIS_LANG_UK = $005D; // Ukrainian
  SIS_LANG_UR = $005E; // Urdu
  { not #95 lang }
  SIS_LANG_VI = $0060; // Vietnamese
  SIS_LANG_SY = $0061; // Welsh
  SIS_LANG_ZU = $0062; // Zulu

  _LANGUAGES_CODES: array[$0000..$0062] of String =
  (
   #0 , 'EN', 'FR', 'GE', 'SP', 'IT', 'SW', 'DA', 'NO', 'FI',
  'AM', 'SW', 'SG', 'PO', 'TU', 'IC', 'RU', 'HU', 'DU', 'BL',
  'AU', 'BG', 'AS', 'NZ', 'IN', 'CS', 'SK', 'PL', 'SL', 'TC',
  'HK', 'ZH', 'JA', 'TH', 'AF', 'SQ', 'AH', 'AR', 'HY', 'TL',
  'BE', 'BN', 'BG', 'MY', 'CA', 'HR', 'CE', 'IE', 'SF', 'ET',
  'FA', 'CF', 'GD', 'KA', 'EL', 'CG', 'GU', 'HE', 'HI', 'IN',
  'GA', 'SZ', 'KN', 'KK', 'KM', 'KO', 'LO', 'LV', 'LT', 'MK',
  'MS', 'ML', 'MR', 'MO', 'MN', 'NN', 'BP', 'PA', 'RO', 'SR',
  'SI', 'SO', 'OS', 'LS', 'SH', 'FS',  #0 , 'TA', 'TE', 'BO',
  'TI', 'CT', 'TK', 'UK', 'UR',  #0 , 'VI', 'SY', 'ZU'
  );

  _LANGUAGES_STRINGS: array[$0000..$0062] of String =
  (
    #0,
    'UK English',
    'French',
    'German',
    'Spanish',
    'Italian',
    'Swedish',
    'Danish',
    'Norwegian',
    'Finnish',
    'American English',
    'Swiss French',
    'Swiss German',
    'Portuguese',
    'Turkish',
    'Icelandic',
    'Russian',
    'Hungarian',
    'Dutch',
    'Belgian Flemish',
    'Australian English',
    'Belgian French',
    'Austrian German',
    'New Zealand English',
    'International French',
    'Czech',
    'Slovak',
    'Polish',
    'Slovenian',
    'Taiwan Chinese',
    'Hong Kong Chinese',
    'PRC Chinese',
    'Japanese',
    'Thai',
    'Afrikaans',
    'Albanian',
    'Amharic',
    'Arabic',
    'Armenian',
    'Tagalog',
    'Belarussian',
    'Bengali',
    'Bulgarian',
    'Burmese',
    'Catalan',
    'Croatian',
    'Canadian English',
    'International English',
    'South African English',
    'Estonian',
    'Farsi',
    'Canadian French',
    'Scots Gaelic',
    'Georgian',
    'Greek',
    'Cyprus Greek',
    'Gujarati',
    'Hebrew',
    'Hindi',
    'Indonesian',
    'Irish',
    'Swiss Italian',
    'Kannada',
    'Kazakh',
    'Khmer',
    'Korean',
    'Laothian',
    'Latvian',
    'Lithuanian',
    'Macedonian',
    'Malay',
    'Malayalam',
    'Marathi',
    'Moldavian',
    'Mongolian',
    'Norwegian-Nynorsk',
    'Brazilian Portuguese',
    'Punjabi',
    'Romanian',
    'Serbian',
    'Sinhalese',
    'Somali',
    'International Spanish',
    'Latin American Spanish',
    'Swahili',
    'Finland Swedish',
    #0,
    'Tamil',
    'Telugu',
    'Tibetan',
    'Tigrinya',
    'Cyprus Turkish',
    'Turkmen',
    'Ukrainian',
    'Urdu',
    #0,
    'Vietnamese',
    'Welsh',
    'Zulu'
  );

type
  _FILE_HEADER = packed record
    FileRecType: DWORD;
    FileType: DWORD;
    FileDetails: DWORD;
    SrcNameLen: DWORD;
    lpSrcName: DWORD;
    DestNameLen: DWORD;
    lpDestName: DWORD;
    FileSize: DWORD;
    lpFile: DWORD;
    OriginalFileSize: DWORD;
    MIMELength: DWORD;
    lpMIMEType: DWORD;
  end;
  TFileHeader = _FILE_HEADER;
  PTFileHeader = ^TFileHeader;

const
  { File Record Type }
  SIS_FILE_REC_SIMPLEFILE = $00000000;
  SIS_FILE_REC_MULTMANGUAGEFILE = $00000001;
  SIS_FILE_REC_OPTIONS = $00000002;
  SIS_FILE_REC_IFLINE = $00000003;
  SIS_FILE_REC_ELSEIFLINE = $00000004;
  SIS_FILE_REC_ELSELINE = $00000005;
  SIS_FILE_REC_ENDIFLINE = $00000006;

  { File type }
  SIS_FILE_TYPE_FF = $00; // FILE	Standard file (the default)
  SIS_FILE_TYPE_FT = $01; // FILETEXT	Text file to display during installation
  SIS_FILE_TYPE_SC = $02; // SIS component file
  SIS_FILE_TYPE_FR = $03; // FILERUN	File to be run during installation and/or removal
  SIS_FILE_TYPE_FN = $04; // FILENULL	File does not yet exist, but will be created when the application is run
  SIS_FILE_TYPE_FM = $05; // FILEMIME	Open file
  
      { FF -> FileDetails = 0x00000000 }
      { FT }
      SIS_FILE_DETAILS_TC = $0000; // TEXTCONTINUE	Continue (continue installation)
      SIS_FILE_DETAILS_TS = $0001; // TEXTSKIP	Yes (continue installation), No (skip next file)
      SIS_FILE_DETAILS_TA = $0002; // TEXTABORT	Yes (continue installation), No (abort installation)
      SIS_FILE_DETAILS_TE = $0003; // TEXTEXIT	Yes (continue installation), No (abort and undo installation)
      { SC -> FileDetails = 0x00000000 }
      { FR }
      SIS_FILE_DETAILS_RI = $0000; // RUNINSTALL	Run during installation only
      SIS_FILE_DETAILS_RR = $0001; // RUNREMOVE	Run during removal only
      SIS_FILE_DETAILS_RB = $0002; // RUNBOTH	Run during both installation and removal
      SIS_FILE_DETAILS_RE = $0100; // RUNSENDEND	Close when installation complete
      SIS_FILE_DETAILS_RW = $0200; // Wait until closed before continuing
      { FN -> FileDetails = 0x00000000  }
      { FM }
      SIS_FILE_DETAILS_OC = $0100; // Close when installation complete
      SIS_FILE_DETAILS_OW = $0200; // Wait until closed before continuing

type
  _REQUISITE = packed record
    UID: DWORD;
    MajorVer: WORD;
    MinorVer: WORD;
    Variant: DWORD;
    NameLen: DWORD;
    lpName: DWORD;
  end;
  TRequisite = _REQUISITE;
  PTRequisite = ^TRequisite;

  _COMPONENT_NAME = packed record
    Length: DWORD;
    lpName: DWORD;
  end;
  TComponentName = _COMPONENT_NAME;
  PTComponentName = ^TComponentName;

  _CAPABILITIES = packed record
    KEY: DWORD;
    VALUE: DWORD;
  end;
  TCapabilities = _CAPABILITIES;
  PTCapabilities = ^TCapabilities;

  _CERTIFICATE = packed record
    Year: WORD;
    Month: WORD;
    Day: WORD;
    Hour: WORD;
    Minute: WORD;
    Second: WORD;
    Count: WORD;
  end;
  TCertificate = _CERTIFICATE;
  PTCertificate = ^TCertificate;

function LockSis(lpFileSis: String): HSIS; stdcall;
function UnLockSis(hSIS: HSIS): LONGBOOL; stdcall;
function ExtractSis(hSIS: HSIS; lpDir: String; Flags: UINT): LONGBOOL; stdcall;
function GetSisName(hSIS: HSIS): String; stdcall;
function GetSisHandleFile(hSIS: HSIS): HFILE; stdcall;
function GetSisVersions(hSIS: HSIS): LongWord; stdcall;
function GetSisInstalationDrive(hSIS: HSIS): WORD; stdcall;
function GetSisUID1(hSIS: HSIS): DWORD; stdcall;
function GetSisUID2(hSIS: HSIS): DWORD; stdcall;
function GetSisUID3(hSIS: HSIS): DWORD; stdcall;
function GetSisUID4(hSIS: HSIS): DWORD; stdcall;
function GetSisCountRequisite(hSIS: HSIS): WORD; stdcall;
function GetSisRequisite(hSIS: HSIS; Index: UINT): PTRequisite; stdcall;
function GetSisRequisiteStr(hSIS: HSIS; Index: UINT): String; stdcall;
function GetSisCountCapabilities(hSIS: HSIS): WORD; stdcall;
function GetSisCapabilities(hSIS: HSIS; Index: UINT): PTCapabilities; stdcall;
function GetSisCountLanguages(hSIS: HSIS): UINT; stdcall;
function GetSisLanguage(hSIS: HSIS; Index: UINT): PTLanguage; stdcall;
function GetSisLanguageStr(hSIS: HSIS; Index: UINT): String; stdcall;
function GetSisLanguageStrCode(hSIS: HSIS; Index: UINT): String; stdcall;
function GetSisCountFiles(hSIS: HSIS): UINT; stdcall;
function GetSisFileInfo(hSIS: HSIS; Index: UINT): PTFileHeader; stdcall;
function GetSisFileDestName(hSIS: HSIS; Index: UINT): String; stdcall;
function GetSisFileSrcName(hSIS: HSIS; Index: UINT): String; stdcall;
function GetSisPkg(hSIS: HSIS; lpFileName, lpDefaultDir: String; uFlag: DWORD): LONGBOOL; stdcall;
function SetSisProcessEvent(hSIS: HMSIS; lpFunction: Pointer): LONGBOOL; stdcall;
function GetSisProcessEvent(hSIS: HMSIS): Pointer; stdcall;

function CreateSis: HMSIS; stdcall;
function FreeSis(hMSIS: HMSIS): LONGBOOL; stdcall;
function BuildSis(hMSIS: HMSIS; lpName: String): LONGBOOL; stdcall;
function SetMSisPkg(hMSIS: HMSIS; lpFileName: String): LONGBOOL; stdcall;
function SetMSisName(hMSIS: HMSIS; lpName: String): LONGBOOL; stdcall;
function GetMSisName(hMSIS: HMSIS): String; stdcall;
function GetMSisUID1(hMSIS: HSIS): DWORD; stdcall;
function GetMSisUID2(hMSIS: HSIS): DWORD; stdcall;
function GetMSisUID3(hMSIS: HSIS): DWORD; stdcall;
function GetMSisUID4(hMSIS: HSIS): DWORD; stdcall;
function SetMSisUID1(hMSIS: HSIS; Value: UINT): LONGBOOL; stdcall;
function SetMSisUID2(hMSIS: HSIS; Value: UINT): LONGBOOL; stdcall;
function SetMSisUID3(hMSIS: HSIS; Value: UINT): LONGBOOL; stdcall;
function GetMSisInstalationDrive(hMSIS: HMSIS): CHAR; stdcall;
function SetMSisInstalationDrive(hMSIS: HMSIS; Drive: CHAR): LONGBOOL; stdcall;
function GetMSisVersions(hMSIS: HMSIS): LongWord; stdcall;
function SetMSisVersions(hMSIS: HMSIS; Value: LongWord): LONGBOOL; stdcall;
function SetMSisCountLanguages(hMSIS: HMSIS; Value: UINT): LONGBOOL; stdcall;
function GetMSisCountLanguages(hMSIS: HMSIS): UINT; stdcall;
function SetMSisLanguage(hMSIS: HMSIS; Index: UINT; lpLanguage: _LANGUAGE): LONGBOOL; stdcall;
function GetMSisLanguage(hMSIS: HMSIS; Index: UINT): PTLanguage; stdcall;
function GetMSisLanguageStr(hMSIS: HMSIS; Index: UINT): String; stdcall;
function GetMSisLanguageStrCode(hMSIS: HMSIS; Index: UINT): String; stdcall;
function SetMSisCountFiles(hMSIS: HMSIS; Value: UINT): LONGBOOL; stdcall;
function GetMSisCountFiles(hMSIS: HMSIS): UINT; stdcall;
function SetMSisFile(hMSIS: HMSIS; Index: UINT; lpFile: _FILE_HEADER): LONGBOOL; stdcall;
function GetMSisFile(hMSIS: HMSIS; Index: UINT): PTFileHeader; stdcall;
function SetMSisFileDestName(hMSIS: HMSIS; Index: UINT; lpName: String): LONGBOOL; stdcall;
function GetMSisFileDestName(hMSIS: HMSIS; Index: UINT): String; stdcall;
function SetMSisFileSrcName(hMSIS: HMSIS; Index: UINT; lpName: String): LONGBOOL; stdcall;
function GetMSisFileSrcName(hMSIS: HMSIS; Index: UINT): String; stdcall;
function SetMSisCountRequisite(hMSIS: HMSIS; Value: UINT): LONGBOOL; stdcall;
function GetMSisCountRequisite(hMSIS: HMSIS): UINT; stdcall;
function SetMSisRequisite(hMSIS: HMSIS; Index: UINT; lpRequisite: _REQUISITE): LONGBOOL; stdcall;
function GetMSisRequisite(hMSIS: HMSIS; Index: UINT): PTRequisite; stdcall;
function SetMSisRequisiteName(hMSIS: HMSIS; Index: UINT; lpName: String): LONGBOOL; stdcall;
function GetMSisRequisiteName(hMSIS: HMSIS; Index: UINT): String; stdcall;
function SetMSisCountCapabilities(hMSIS: HMSIS; Value: UINT): LONGBOOL; stdcall;
function GetMSisCountCapabilities(hMSIS: HMSIS): UINT; stdcall;
function SetMSisCapabilities(hMSIS: HMSIS; Index: UINT; lpCapabilities: _CAPABILITIES): LONGBOOL; stdcall;
function GetMSisCapabilities(hMSIS: HMSIS; Index: UINT): PTCapabilities; stdcall;
function SetMSisProcessEvent(hMSIS: HMSIS; lpFunction: Pointer): LONGBOOL; stdcall;
function GetMSisProcessEvent(hMSIS: HMSIS): Pointer; stdcall;

function MakeIDLanguage(lpLang: String): WORD; stdcall;
function MakeSisVersions(Major, Minor: WORD): LongWord; stdcall;
function GetSisMajorVersion(Versions: LongWord): WORD; stdcall;
function GetSisMinorVersion(Versions: LongWord): WORD; stdcall;

implementation

procedure DoSisEvent(SisHandle, ID: UINT; lParam: LPARAM; wParam: WPARAM;
  SisEvent: _SIS_EVENT);
begin
  if (Assigned(SisEvent)) then
    SisEvent(SisHandle, ID, lParam, wParam);
end;

function StrToUInt(Value:String; var iResult): Boolean;
var
  code : integer;
begin
  val(value,uint(iresult),code);
  result := code = 0;
end;

function MakeIDLanguage;
var
  i: WORD;
begin
  Result := 0;
  for i:= $0000 to $0062 do
    if (_LANGUAGES_CODES[i] = lpLang) then
    begin
      Result := i;
      break;
    end;
end;

function MakeSisVersions(Major, Minor: WORD): LongWord;
begin
  Result := MakeLong(Major, Minor);
end;

function GetSisMajorVersion(Versions: LongWord): WORD;
begin
  Result := LOWORD(Versions);
end;

function GetSisMinorVersion(Versions: LongWord): WORD;
begin
  Result := HIWORD(Versions);
end;


type
{ HMSIS }
{
  File Header
  Language Records
  File Records
  Requisite Records
  Component Name Record
  Capabilities Record
  Certificates Record
  Resource Data
}
  _DATA_HANDLE_MSIS = packed record
    SIS_FILE_HEADER: _SIS_FILE_HEADER;
    LANGUAGES: array of _LANGUAGE;
    FILES: array of _FILE_HEADER;
    REQUISITES: array of _REQUISITE;
    SISNAME: String;
    COMPONENT_NAME: _COMPONENT_NAME;
    CAPABILITIES: array of _CAPABILITIES;
    FILESRCNAMES: array of String;
    FILEDESTNAMES: array of String;
    REQUISITENAMES: array of String;
    PROCESS_EVENT: _SIS_EVENT;
  end;
  p_DATA_HANDLE_MSIS = ^_DATA_HANDLE_MSIS;

function CreateSis;
var
  pHMSIS: Pointer;
begin
  GetMem(pHMSIS, SIZEOF(_DATA_HANDLE_MSIS));
  with (p_DATA_HANDLE_MSIS(pHMSIS)^), SIS_FILE_HEADER do
  begin
    PROCESS_EVENT := NIL;
    UID1 := EPOC_UID_DEF_APP;
    UID2 := EPOC_UID_REL_6;
    UID3 := EPOC_UID_ALWAYS;
    InstallationDrive := EPOC_INS_DRIVE_REL_6;
    InstallerVersion := EPOC_INS_VER_REL_6;
    Options := SIS_OPT_IU;
    sisType := SIS_TYPE_SA;
  end;
  Result := HMSIS(pHMSIS);
end;

function FreeSis;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    DoSisEvent(hMSIS, SIS_PROC_TYPE_FREE, 0, 0, PROCESS_EVENT);
    SetLength(LANGUAGES, 0);
    SetLength(REQUISITES, 0);
    SetLength(CAPABILITIES, 0);
    SetLength(FILEDESTNAMES, 0);
    SetLength(FILESRCNAMES, 0);
    FreeMem(Pointer(hMSIS));
    Result := TRUE;
  end else
    Result := FALSE;
end;

function SetMSisProcessEvent;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    PROCESS_EVENT := _SIS_EVENT(lpFunction^);
    Result := TRUE;
  end;
end;

function GetMSisProcessEvent;
begin
  Result := NIL;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := @PROCESS_EVENT;
  end;
end;

function SetMSisName;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SISNAME := lpName;
    COMPONENT_NAME.Length := Length(SISNAME) * 2;
    Result := TRUE;
  end;
end;

function GetMSisName;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SISNAME;
  end;
end;

function SetMSisUID1;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SIS_FILE_HEADER.UID1 := Value;
    Result := TRUE;
  end;
end;

function SetMSisUID2;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SIS_FILE_HEADER.UID2 := Value;
    Result := TRUE;
  end;
end;

function SetMSisUID3;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SIS_FILE_HEADER.UID3 := Value;
    Result := TRUE;
  end;
end;

function GetMSisUID1;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.UID1;
  end;
end;

function GetMSisUID2;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.UID2;
  end;
end;

function GetMSisUID3;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.UID3;
  end;
end;

function GetMSisUID4;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.UID4;
  end;
end;

function GetMSisInstalationDrive;
begin
  Result := #0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := chr(SIS_FILE_HEADER.InstallationDrive);
  end;
end;

function SetMSisInstalationDrive;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SIS_FILE_HEADER.InstallationDrive := ord(Drive);
    Result := TRUE;
  end;
end;

function SetMSisVersions;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    SIS_FILE_HEADER.MajorVersion := GetSisMajorVersion(Value);
    SIS_FILE_HEADER.MinorVersion := GetSisMinorVersion(Value);
    Result := TRUE;
  end;
end;

function GetMSisVersions;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := MakeSisVersions(SIS_FILE_HEADER.MajorVersion, SIS_FILE_HEADER.MinorVersion);
  end;
end;

function SetMSisCountLanguages;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    if (Value <> SIS_FILE_HEADER.NumberOfLanguages) and (Value>=1) then
    begin
      SetLength(LANGUAGES, Value);
      SIS_FILE_HEADER.NumberOfLanguages := Value;
      Result := TRUE;
    end;
  end;
end;

function GetMSisCountLanguages;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.NumberOfLanguages;
  end;
end;

function SetMSisLanguage;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfLanguages) then
  begin
    LANGUAGES[Index] := lpLanguage;
    Result := TRUE;
  end;
end;

function GetMSisLanguage;
begin
  Result := NIL;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfLanguages) then
    Result := @LANGUAGES[Index];
end;

function GetMSisLanguageStr;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  Result := _LANGUAGES_STRINGS[LANGUAGES[Index].ID];
end;

function GetMSisLanguageStrCode;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  Result := _LANGUAGES_CODES[LANGUAGES[Index].ID];
end;

function SetMSisCountFiles;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    if (Value <> SIS_FILE_HEADER.NumberOfFiles) and (Value>=1) then
    begin
      SetLength(FILES, Value);
      SetLength(FILEDESTNAMES, Value);
      SetLength(FILESRCNAMES, Value);
      SIS_FILE_HEADER.NumberOfFiles := Value;
      Result := TRUE;
    end;
  end;
end;

function GetMSisCountFiles;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.NumberOfFiles;
  end;
end;

function SetMSisFile;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    FILES[index] := lpFile;
    Result := TRUE;
  end;
end;

function SetMSisFileSrcName;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    FILESRCNAMES[index] := lpName;
    FILES[index].SrcNameLen := Length(FILESRCNAMES[index]) * 2;
    Result := TRUE;
  end;
end;

function SetMSisFileDestName;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    FILEDESTNAMES[index] := lpName;
    FILES[index].DestNameLen := Length(FILEDESTNAMES[index]) * 2;
    Result := TRUE;
  end;
end;

function GetMSisFileSrcName;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    Result := FILESRCNAMES[Index];
  end;
end;

function GetMSisFileDestName;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    Result := FILEDESTNAMES[Index];
  end;
end;

function GetMSisFile;
begin
  Result := NIL;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    Result := @FILES[Index];
  end;
end;

function SetMSisCountRequisite;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    if (Value <> SIS_FILE_HEADER.NumberOfRequisites) and (Value >= 1) then
    begin
      SetLength(REQUISITES, Value);
      SetLength(REQUISITENAMES, Value);
      SIS_FILE_HEADER.NumberOfRequisites := Value;
      Result := TRUE;
    end;
  end;
end;            

function GetMSisCountRequisite;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.NumberOfRequisites;
  end;
end;

function SetMSisRequisite;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfRequisites) then
  begin
    REQUISITES[index] := lpRequisite;
    Result := TRUE;
  end;
end;

function GetMSisRequisite;
begin
  Result := NIL;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfRequisites) then
  begin
    Result := @REQUISITES[Index];
  end;
end;

function SetMSisRequisiteName;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    REQUISITENAMES[index] := lpName;
    REQUISITES[index].NameLen := Length(REQUISITENAMES[index]) * 2;
    Result := TRUE;
  end;
end;

function GetMSisRequisiteName;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfFiles) then
  begin
    Result := REQUISITENAMES[index];
  end;
end;

function SetMSisCountCapabilities;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    if (Value <> SIS_FILE_HEADER.NumberOfCapabilities) then
    begin
      SetLength(CAPABILITIES, Value);
      SIS_FILE_HEADER.NumberOfCapabilities := Value;
      Result := TRUE;
    end;
  end;
end;

function GetMSisCountCapabilities;
begin
  Result := 0;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Result := SIS_FILE_HEADER.NumberOfCapabilities;
  end;
end;

function SetMSisCapabilities;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfCapabilities) then
  begin
    CAPABILITIES[index] := lpCapabilities;
    Result := TRUE;
  end;
end;

function GetMSisCapabilities;
begin
  Result := NIL;
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (Index < SIS_FILE_HEADER.NumberOfCapabilities) then
  begin
    Result := @CAPABILITIES[Index];
  end;
end;
                      
function SetMSisPkg;

procedure ReverseFileData;
var
  i, m: WORD;
  lpFile: TFileHeader;
  lpName: String;
begin
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  if (SIS_FILE_HEADER.NumberOfFiles>0) then
  begin
    m := (SIS_FILE_HEADER.NumberOfFiles-1);
    for i:= 0 to (m div 2) do
    begin
      lpFile := FILES[i];
      FILES[i] := FILES[m-i];
      FILES[m-i] := lpFile;
      lpName := FILESRCNAMES[i];
      FILESRCNAMES[i] := FILESRCNAMES[m-i];
      FILESRCNAMES[m-i] := lpName;
      lpName := FILEDESTNAMES[i];
      FILEDESTNAMES[i] := FILEDESTNAMES[m-i];
      FILEDESTNAMES[m-i] := lpName;
    end;
  end;
end;

var
  inpFILE: HFILE;
  i, n, len, start: Integer;
  lpReOpenBuff: _OFSTRUCT;
  lpLinePkg: String;
  lpFileFlags: array[0..2] of String;
  LenLinePkg: Integer;
  Drive: CHAR;
begin
  Result := FALSE;
  if (hMSIS <> INVALID_SIS_HANDLE) and (FileExists(lpFileName)) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    Drive := '!';
    inpFILE := OpenFile(PCHAR(lpFileName), lpReOpenBuff, OF_READ);
    while (FileReadLine(inpFILE, lpLinePkg)) do
    begin
      LenLinePkg := Length(lpLinePkg);
      if (LenLinePkg > 0) and (lpLinePkg[1] <> ';') then
      begin

        if (lpLinePkg[1]='&') then
        begin
          start := 2;
          for i:= 1 to LenLinePkg do
            if (lpLinePkg[i] = ',') or (i = LenLinePkg) then
            begin
              len := i - start;
              if (i = LenLinePkg) then len := len + 1;
              SIS_FILE_HEADER.NumberOfLanguages := SIS_FILE_HEADER.NumberOfLanguages + 1;
              SetLength(LANGUAGES, SIS_FILE_HEADER.NumberOfLanguages);
              LANGUAGES[SIS_FILE_HEADER.NumberOfLanguages-1].ID :=
                MakeIDLanguage(Copy(lpLinePkg, Start, Len));
              start := i + 1;
            end;
        end else
        if (lpLinePkg[1]='#') then
        begin
          start := 4;
          len := LCharPos('"', lpLinePkg, start) - start;
          COMPONENT_NAME.Length := len * 2;
          SISNAME := Copy(lpLinePkg, Start, Len);
                                          
          start := LCharPos('(',lpLinePkg, 1) + 3;
          len := LCharPos(')',lpLinePkg,start) - start;
          StrToUInt(('$'+Copy(lpLinePkg, Start, Len)), SIS_FILE_HEADER.UID1);

          start := LCharPos(',',lpLinePkg, start) + 1;
          len := LCharPos(',',lpLinePkg,start) - start;
          StrToUInt(Copy(lpLinePkg, Start, Len), SIS_FILE_HEADER.MajorVersion);
          
          start := LCharPos(',',lpLinePkg, start) + 1;
          len := LCharPos(',',lpLinePkg, start) - start;
          StrToUInt(Copy(lpLinePkg, Start, Len), SIS_FILE_HEADER.MinorVersion);

          start := LCharPos(',',lpLinePkg, start);
          len := Length(lpLinePkg) - start;
          StrToUInt(Copy(lpLinePkg, (Start+1), Len), SIS_FILE_HEADER.Variant);
        end else
        if (lpLinePkg[1]='(') then
        begin
          SIS_FILE_HEADER.NumberOfRequisites := SIS_FILE_HEADER.NumberOfRequisites + 1;
          SetLength(REQUISITES, SIS_FILE_HEADER.NumberOfRequisites);
          SetLength(REQUISITENAMES, SIS_FILE_HEADER.NumberOfRequisites);
          
          start := LCharPos('(',lpLinePkg, 1) + 3;
          len := LCharPos(')',lpLinePkg,start)-start;
          StrToUInt(('$'+Copy(lpLinePkg, Start, Len)),
            REQUISITES[SIS_FILE_HEADER.NumberOfRequisites-1].UID);

          start := LCharPos(',',lpLinePkg, 1) + 1;
          len := LCharPos(',',lpLinePkg,start)-start;
          StrToUInt(Copy(lpLinePkg, Start, Len),
            REQUISITES[SIS_FILE_HEADER.NumberOfRequisites-1].MajorVer);

          start := LCharPos(',',lpLinePkg,start)+1;
          len := LCharPos(',',lpLinePkg,start)-start;
          StrToUInt(Copy(lpLinePkg, Start, Len),
            REQUISITES[SIS_FILE_HEADER.NumberOfRequisites-1].MinorVer);

          start := LCharPos(',',lpLinePkg,start)+1;
          len := LCharPos(',',lpLinePkg,start)-start;
          StrToUInt(Copy(lpLinePkg, Start, Len),
            REQUISITES[SIS_FILE_HEADER.NumberOfRequisites-1].Variant);

          start := LCharPos('{',lpLinePkg,start)+2;
          len := LCharPos('"',lpLinePkg,start)-start;
          REQUISITENAMES[SIS_FILE_HEADER.NumberOfRequisites-1] := Copy(lpLinePkg, Start, Len);
          REQUISITES[SIS_FILE_HEADER.NumberOfRequisites-1].NameLen :=
            Length(REQUISITENAMES[SIS_FILE_HEADER.NumberOfRequisites-1])*2;
        end else
        if (lpLinePkg[1]='"') then
        begin
          SIS_FILE_HEADER.NumberOfFiles := SIS_FILE_HEADER.NumberOfFiles + 1;
          SetLength(FILES, SIS_FILE_HEADER.NumberOfFiles);
          SetLength(FILEDESTNAMES, SIS_FILE_HEADER.NumberOfFiles);
          SetLength(FILESRCNAMES, SIS_FILE_HEADER.NumberOfFiles);

          start := LCharPos('"',lpLinePkg,0)+1;
          len := LCharPos('"',lpLinePkg,start)-start;
          FILESRCNAMES[SIS_FILE_HEADER.NumberOfFiles-1] := Copy(lpLinePkg, Start, Len);
          FILES[SIS_FILE_HEADER.NumberOfFiles-1].SrcNameLen :=
            Length(FILESRCNAMES[SIS_FILE_HEADER.NumberOfFiles-1]) * 2;

          start := LCharPos('"',lpLinePkg,start+len)+3;
          len := LCharPos('"',lpLinePkg,start)-start;
          FILEDESTNAMES[SIS_FILE_HEADER.NumberOfFiles-1] := Copy(lpLinePkg, Start, Len);
          FILES[SIS_FILE_HEADER.NumberOfFiles-1].DestNameLen :=
            Length(FILEDESTNAMES[SIS_FILE_HEADER.NumberOfFiles-1]) * 2;

          if (FILES[SIS_FILE_HEADER.NumberOfFiles-1].DestNameLen>0) then
            Drive := FILEDESTNAMES[SIS_FILE_HEADER.NumberOfFiles-1][1];

          lpFileFlags[0] := #0;
          lpFileFlags[1] := #0;
          lpFileFlags[2] := #0;
          for n:= 0 to 2 do
          begin
            start := LCharPos(',',lpLinePkg,start+len)+1;
            len := LCharPos(',',lpLinePkg,start);
            if (len>0) then
            begin
              len := len - start;
              lpFileFlags[n] := Copy(lpLinePkg, Start, Len);
            end else
            begin
              len := 2;
              lpFileFlags[n] := Copy(lpLinePkg, Start, Len);
              break;
            end;
          end;
          
          with (FILES[SIS_FILE_HEADER.NumberOfFiles-1]) do
          begin
            if (lpFileFlags[0]='FF') then
              FileType := SIS_FILE_TYPE_FF
            else
            if (lpFileFlags[0]='FT') then
            begin
              FileType := SIS_FILE_TYPE_FT;
              if (lpFileFlags[1]='TC') then
                FileDetails := SIS_FILE_DETAILS_TC
              else
              if (lpFileFlags[1]='TS') then
                FileDetails := SIS_FILE_DETAILS_TS
              else
              if (lpFileFlags[1]='TA') then
                FileDetails := SIS_FILE_DETAILS_TA
              else
              if (lpFileFlags[1]='TE') then
                FileDetails := SIS_FILE_DETAILS_TE;
            end else
            if (lpFileFlags[0]='FR') then
            begin
              FileType := SIS_FILE_TYPE_FR;
              if (lpFileFlags[1]='RI') then
                FileDetails := SIS_FILE_DETAILS_RI
              else
              if (lpFileFlags[1]='RR') then
                FileDetails := SIS_FILE_DETAILS_RR
              else
              if (lpFileFlags[1]='RB') then
                FileDetails := SIS_FILE_DETAILS_RB;
              if (lpFileFlags[2]='RE') then
                FileDetails := FileDetails or SIS_FILE_DETAILS_RE
              else
              if (lpFileFlags[2]='RW') then
                FileDetails := FileDetails or SIS_FILE_DETAILS_RW;
            end else
            if (lpFileFlags[0]='FN') then
              FileType := SIS_FILE_TYPE_FN
            else
            if (lpFileFlags[0]='FM') then
            begin
              FileType := SIS_FILE_TYPE_FM;
              if (lpFileFlags[1]='OC') then
                FileDetails := SIS_FILE_DETAILS_OC
              else
              if (lpFileFlags[1]='OW') then
                FileDetails := SIS_FILE_DETAILS_OW;
            end;
          end;
        end;     

      end;
    end;
    CloseHandle(inpFILE);
    ReverseFileData;
    SetMSisInstalationDrive(hMSIS, Drive);
  end;
end;

function BuildSis;
var
  HDFILE, inpFILE: HFILE;
  lpReOpenBuff: _OFSTRUCT;
  i: UINT;
  lpNumberOfBytesWritten, lpNumberOfBytesRead: DWORD;
  ResourceDataOffset: DWORD;
  inpBuf, OutBuf: Pointer;
  inpBytes, outBytes: Integer;
begin
  Result := FALSE;    
  if (hMSIS <> INVALID_SIS_HANDLE) then
  with (p_DATA_HANDLE_MSIS(hMSIS)^) do
  begin
    DeleteFile(LPCSTR(lpName));
    HDFILE := OpenFile(LPCSTR(lpName), lpReOpenBuff, OF_CREATE);
    if (HDFILE <> HFILE_ERROR) then
    begin
      DoSisEvent(hMSIS, SIS_PROC_TYPE_BUILD, Integer(lpName), 0, PROCESS_EVENT);
      
      ResourceDataOffset :=
        SIZEOF(_SIS_FILE_HEADER)+
        SIZEOF(_LANGUAGE)*SIS_FILE_HEADER.NumberOfLanguages+
        SIZEOF(_FILE_HEADER)*SIS_FILE_HEADER.NumberOfFiles+
        SIZEOF(_REQUISITE)*SIS_FILE_HEADER.NumberOfRequisites+
        SIZEOF(_COMPONENT_NAME)+
        SIZEOF(_CAPABILITIES)*SIS_FILE_HEADER.NumberOfCapabilities{+
        SIZEOF(_CERTIFICATES)};

      // seek on resource data
      SetFilePointer(HDFILE, ResourceDataOffset, NIL, FILE_BEGIN);
      // write name files
      if (SIS_FILE_HEADER.NumberOfFiles>0) then
      for i:= 0 to (SIS_FILE_HEADER.NumberOfFiles-1) do
      begin
        FILES[i].lpSrcName := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
        WriteUnicode(HDFILE, FILESRCNAMES[i]);
        FILES[i].lpDestName := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
        WriteUnicode(HDFILE, FILEDESTNAMES[i]);
      end;
      // write requisties name
      if (SIS_FILE_HEADER.NumberOfRequisites>0) then
      for i:= 0 to (SIS_FILE_HEADER.NumberOfRequisites-1) do
      begin  
        REQUISITES[i].lpName := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
        WriteUnicode(HDFILE, REQUISITENAMES[i]);
      end;
      // write sis name                         
      if (COMPONENT_NAME.Length>0) then
      begin
        COMPONENT_NAME.lpName := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
        WriteUnicode(HDFILE, SISNAME);
      end;

      SIS_FILE_HEADER.spMaximumInstalledSpace := 0;
      // write files data
      if (SIS_FILE_HEADER.NumberOfFiles>0) then
      for i:= 0 to (SIS_FILE_HEADER.NumberOfFiles-1) do
      begin
        lpNumberOfBytesWritten := 0;
        FILES[i].lpFile := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);

        inpFILE := OpenFile(LPCSTR(FILESRCNAMES[i]), lpReOpenBuff, OF_READWRITE);
        inpBytes := GetFileSize(inpFILE, NIL);
        GetMem(inpBuf, inpBytes);
        ReadFile(inpFILE, inpBuf^, inpBytes, lpNumberOfBytesRead, NIL);
        zlib.CompressBuf(inpBuf, lpNumberOfBytesRead, outBuf, outBytes);
        WriteFile(HDFILE, outBuf^, outBytes, lpNumberOfBytesWritten, NIL);
        FreeMem(outBuf);
        FreeMem(inpBuf);
        CloseHandle(inpFILE);
    
        FILES[i].OriginalFileSize := inpBytes;
        FILES[i].FileSize := lpNumberOfBytesWritten;
        
        SIS_FILE_HEADER.spMaximumInstalledSpace :=
          SIS_FILE_HEADER.spMaximumInstalledSpace + FILES[i].OriginalFileSize;

        DoSisEvent(hMSIS, SIS_PROC_TYPE_FILE, Integer(FILESRCNAMES[i]), 0, PROCESS_EVENT);
      end;

      // write Languages
      if (SIS_FILE_HEADER.NumberOfLanguages>0) then
      begin
      SIS_FILE_HEADER.lpLanguages := SetFilePointer(HDFILE, SIZEOF(_SIS_FILE_HEADER), NIL, FILE_BEGIN);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfLanguages-1) do
        WriteFile(HDFILE, LANGUAGES[i], SIZEOF(_LANGUAGE), lpNumberOfBytesWritten, NIL);
      end;

      // write file header
      if (SIS_FILE_HEADER.NumberOfFiles>0) then
      begin
      SIS_FILE_HEADER.lpFiles := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfFiles-1) do
      begin
        if (FILES[i].MIMELength=0) then
        begin                  
          if (i<DWORD(SIS_FILE_HEADER.NumberOfFiles-1)) then
            FILES[i].lpMIMEType := FILES[i+1].lpSrcName
          else
          if (SIS_FILE_HEADER.NumberOfRequisites>0) then
            FILES[i].lpMIMEType := REQUISITES[0].lpName
          else
            FILES[i].lpMIMEType := COMPONENT_NAME.lpName;
        end;
        WriteFile(HDFILE, FILES[i], SIZEOF(_FILE_HEADER), lpNumberOfBytesWritten, NIL);
      end;
      end;
      
      // write Requisites
      if (SIS_FILE_HEADER.NumberOfRequisites>0) then
      begin
      SIS_FILE_HEADER.lpRequisites := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfRequisites-1) do
        WriteFile(HDFILE, REQUISITES[i], SIZEOF(_REQUISITE), lpNumberOfBytesWritten, NIL);
      end;
      
      // write Component
      SIS_FILE_HEADER.lpComponentName := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
      WriteFile(HDFILE, COMPONENT_NAME, SIZEOF(_COMPONENT_NAME), lpNumberOfBytesWritten, NIL);

      // write Capabilities
      if (SIS_FILE_HEADER.NumberOfCapabilities>0) then
      begin
      SIS_FILE_HEADER.lpCapabilities := SetFilePointer(HDFILE, 0, NIL, FILE_CURRENT);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfCapabilities-1) do
        WriteFile(HDFILE, CAPABILITIES[i], SIZEOF(_CAPABILITIES), lpNumberOfBytesWritten, NIL);
      end;

      // ???
      if (SIS_FILE_HEADER.NumberOfRequisites=0) then
        SIS_FILE_HEADER.lpRequisites := SIS_FILE_HEADER.lpComponentName;
      if (SIS_FILE_HEADER.NumberOfCapabilities=0) then
        SIS_FILE_HEADER.lpCapabilities := SIS_FILE_HEADER.lpComponentName;
      
      // write header
      SetFilePointer(HDFILE, 0, NIL, FILE_BEGIN);
      WriteFile(HDFILE, SIS_FILE_HEADER, SIZEOF(_SIS_FILE_HEADER), lpNumberOfBytesWritten, NIL);     

      CloseHandle(HDFILE);
      Result := TRUE;
    end;
  end;
end;


type
{ HSIS }
  _DATA_HANDLE_SIS = packed record
    HANDLE_FILE: HFILE;
    SIS_FILE_HEADER: _SIS_FILE_HEADER;
    FILE_HEADER: _FILE_HEADER;
    SISNAME: String;
    LANGUAGE: _LANGUAGE;
    COMPONENT_NAME: _COMPONENT_NAME;
    REQUISITE: _REQUISITE;
    CAPABILITIES: _CAPABILITIES;
    PROCESS_EVENT: _SIS_EVENT;
  end;
  p_DATA_HANDLE_SIS = ^_DATA_HANDLE_SIS;
  

function LockSis;
var
  lpDATAHSIS: p_DATA_HANDLE_SIS;
  HFL: HFILE;
  HSS: HSIS;
  lpOpenBuff: _OFSTRUCT;
  lpNumberOfBytesRead: DWORD;
begin
  Result := INVALID_SIS_HANDLE;
  HSS := INVALID_SIS_HANDLE;
  if (FileExists(lpFileSis)) then
  begin           
    HFL := OpenFile(LPCSTR(lpFileSis), lpOpenBuff, OF_READWRITE);
    if (HFL <> HFILE_ERROR) then
    begin
      GetMem(lpDATAHSIS, SIZEOF(_DATA_HANDLE_SIS));
      if (lpDATAHSIS <> NIL) then
      with lpDATAHSIS^ do
      begin
        HANDLE_FILE := HFL;
        HSS := HSIS(lpDATAHSIS);

        ReadFile(HFL, SIS_FILE_HEADER, SIZEOF(_SIS_FILE_HEADER), lpNumberOfBytesRead, NIL);

        if (SIS_FILE_HEADER.lpComponentName > 0) then
        begin
          SetFilePointer(HFL, SIS_FILE_HEADER.lpComponentName, NIL, FILE_BEGIN);
          ReadFile(HFL, COMPONENT_NAME, SIZEOF(_COMPONENT_NAME), lpNumberOfBytesRead, NIL);
          if (COMPONENT_NAME.Length > 0) then
          begin
            SetFilePointer(HFL, COMPONENT_NAME.lpName, NIL, FILE_BEGIN);
            SISNAME := ReadUnicode(HFL, COMPONENT_NAME.Length);
          end;
        end;   
      end;
      Result := HSS;
    end;
  end;
end;

function GetSisPkg;
var
  Pkg: HFILE;
  lpReOpenBuff: _OFSTRUCT;
  lpNumberOfBytes: UINT;
  i: WORD;

procedure WriteFileStr(H: HFILE; lpStr: String);
begin
  WriteFile(H, LPCSTR(lpStr)^, Length(lpStr), lpNumberOfBytes, NIL);
end;

begin
  Result := FALSE;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    Pkg := OpenFile(LPCSTR(lpFileName), lpReOpenBuff, OF_READWRITE or OF_CREATE);

    WriteFileStr(Pkg, ';Auto generate PKG file'#13#10);
   // WriteFileStr(Pkg, ';2007 '#$00A9'  .., UKRAINE'#13#10#13#10);

    if (SIS_FILE_HEADER.NumberOfLanguages>0) then
    begin
      WriteFileStr(Pkg, ';Languages'#13#10'&');
      for i:= 0 to (SIS_FILE_HEADER.NumberOfLanguages-1) do
      begin
        WriteFileStr(Pkg, GetSisLanguageStrCode(hSIS, i));
        if (i<(SIS_FILE_HEADER.NumberOfLanguages-1)) then
          WriteFileStr(Pkg, ',');
      end;
      WriteFileStr(Pkg, #13#10#13#10);
    end;

    WriteFileStr(Pkg, ';Header'#13#10'#{"');
    WriteFileStr(Pkg, SISNAME);
    WriteFileStr(Pkg, '"},(0x');
    WriteFileStr(Pkg, inttohex(SIS_FILE_HEADER.UID1, 8));
    WriteFileStr(Pkg, '),');
    WriteFileStr(Pkg, inttostr(SIS_FILE_HEADER.MajorVersion));
    WriteFileStr(Pkg, ',');
    WriteFileStr(Pkg, inttostr(SIS_FILE_HEADER.MinorVersion));
    WriteFileStr(Pkg, ',');
    WriteFileStr(Pkg, inttostr(SIS_FILE_HEADER.Variant));
    WriteFileStr(Pkg, #13#10#13#10);

    if (SIS_FILE_HEADER.NumberOfRequisites>0) then
    begin
      WriteFileStr(Pkg, ';Requisite'#13#10);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfRequisites-1) do
      begin
        GetSisRequisite(hSIS, i);
        WriteFileStr(Pkg, '(0x');
        WriteFileStr(Pkg, inttohex(REQUISITE.UID, 8));
        WriteFileStr(Pkg, '), ');
        WriteFileStr(Pkg, inttostr(REQUISITE.MajorVer));
        WriteFileStr(Pkg, ', ');
        WriteFileStr(Pkg, inttostr(REQUISITE.MinorVer));
        WriteFileStr(Pkg, ', ');
        WriteFileStr(Pkg, inttostr(REQUISITE.Variant));
        WriteFileStr(Pkg, ', {"');
        WriteFileStr(Pkg, GetSisRequisiteStr(hSIS, i));
        WriteFileStr(Pkg, '"}');
        WriteFileStr(Pkg, #13#10);
      end;
      WriteFileStr(Pkg, #13#10);
    end;

    if (SIS_FILE_HEADER.NumberOfFiles>0) then
    begin
      WriteFileStr(Pkg, ';Files'#13#10);
      for i:= 0 to (SIS_FILE_HEADER.NumberOfFiles-1) do
      with GetSisFileInfo(hSIS, i)^ do
      if (FileType=SIS_FILE_TYPE_SC) then
      begin
        WriteFileStr(Pkg, '@"');
        WriteFileStr(Pkg, lpDefaultDir);
        if (DestNameLen>0) then
          WriteFileStr(Pkg, ExtractFileName(GetSisFileDestName(hSIS, i)))
        else
          WriteFileStr(Pkg, ExtractFileName(GetSisFileSrcName(hSIS, i)));
        WriteFileStr(Pkg, '",(0x');
        WriteFileStr(Pkg, inttohex(FileDetails, 8));
        WriteFileStr(Pkg, ')'#13#10);
      end else
      begin
        WriteFileStr(Pkg, '"');
        WriteFileStr(Pkg, lpDefaultDir);

        if (DestNameLen>0) then
        begin
          case uFlag of
            SIS_SAVEDIR:WriteFileStr(Pkg, Copy(GetSisFileDestName(hSIS, i), 4,
              Length(GetSisFileDestName(hSIS, i))));
            SIS_NODIR:WriteFileStr(Pkg, ExtractFileName(GetSisFileDestName(hSIS, i)));
          end;
        end else
          WriteFileStr(Pkg, ExtractFileName(GetSisFileSrcName(hSIS, i)));
        WriteFileStr(Pkg, '"-"');
        WriteFileStr(Pkg, GetSisFileDestName(hSIS, i));
        WriteFileStr(Pkg, '"');
        with GetSisFileInfo(hSIS, i)^ do
        begin
          case FileType of
            SIS_FILE_TYPE_FF:WriteFileStr(Pkg, ',FF');
            SIS_FILE_TYPE_FT:
            begin
              WriteFileStr(Pkg, ',FT');
              case FileDetails of
                SIS_FILE_DETAILS_TC:WriteFileStr(Pkg, ',TC');
                SIS_FILE_DETAILS_TS:WriteFileStr(Pkg, ',TS');
                SIS_FILE_DETAILS_TA:WriteFileStr(Pkg, ',TA');
                SIS_FILE_DETAILS_TE:WriteFileStr(Pkg, ',TE');
              end;
            end;
            SIS_FILE_TYPE_FR:
            begin                  
              WriteFileStr(Pkg, ',FR');
              case (FileDetails and $FF) of
                SIS_FILE_DETAILS_RI:WriteFileStr(Pkg, ',RI');
                SIS_FILE_DETAILS_RR:WriteFileStr(Pkg, ',RR');
                SIS_FILE_DETAILS_RB:WriteFileStr(Pkg, ',RB');
              end;
              case (FileDetails and $FF00) of
                SIS_FILE_DETAILS_RE:WriteFileStr(Pkg, ',RE');
                SIS_FILE_DETAILS_RW:WriteFileStr(Pkg, ',RW');
              end;
            end;
            SIS_FILE_TYPE_FN:WriteFileStr(Pkg, ',FN');
            SIS_FILE_TYPE_FM:
            begin
              WriteFileStr(Pkg, ',FM');
              case FileDetails of
                SIS_FILE_DETAILS_OC:WriteFileStr(Pkg, ',OC');
                SIS_FILE_DETAILS_OW:WriteFileStr(Pkg, ',OW');
              end;
            end;
          end;
        end;
        WriteFileStr(Pkg, #13#10);
      end;
      WriteFileStr(Pkg, #13#10);
    end;

    CloseHandle(Pkg);
  end;
end;

function UnLockSis;
begin
  Result := FALSE;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    if (HANDLE_FILE <> HFILE_ERROR) then
      CloseHandle(HANDLE_FILE);
    SysFreeMem(Pointer(hSIS));
    Result := TRUE;
  end;
end;

function SetSisProcessEvent;
begin
  Result := FALSE;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    PROCESS_EVENT := _SIS_EVENT(lpFunction^);
    Result := TRUE;
  end;
end;

function GetSisProcessEvent;
begin
  Result := NIL;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    Result := @PROCESS_EVENT;
  end;
end;

function GetSisName;
begin
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SISNAME;
end;

function GetSisCountFiles;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.NumberOfFiles;
end;

function GetSisFileInfo;
var
  lpNumberOfBytesRead: DWORD;
begin
  Result := NIL;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfFiles) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpFiles + (SIS_FILE_HEADER.NumberOfFiles-Index-1)*SIZEOF(_FILE_HEADER),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, FILE_HEADER, SIZEOF(_FILE_HEADER), lpNumberOfBytesRead, NIL);
    Result := @FILE_HEADER;
  end;
end;

function GetSisFileDestName;
var
  lpNumberOfBytesRead: DWORD;
begin
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfFiles) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpFiles + (SIS_FILE_HEADER.NumberOfFiles-Index-1)*SIZEOF(_FILE_HEADER),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, FILE_HEADER, SIZEOF(_FILE_HEADER), lpNumberOfBytesRead, NIL);
    SetFilePointer(HANDLE_FILE, FILE_HEADER.lpDestName, NIL, FILE_BEGIN);
    Result := ReadUnicode(HANDLE_FILE, FILE_HEADER.DestNameLen);
  end;
end;

function GetSisFileSrcName;
var
  lpNumberOfBytesRead: DWORD;
begin
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfFiles) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpFiles + (SIS_FILE_HEADER.NumberOfFiles-Index-1)*SIZEOF(_FILE_HEADER),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, FILE_HEADER, SIZEOF(_FILE_HEADER), lpNumberOfBytesRead, NIL);
    SetFilePointer(HANDLE_FILE, FILE_HEADER.lpSrcName, NIL, FILE_BEGIN);
    Result := ReadUnicode(HANDLE_FILE, FILE_HEADER.SrcNameLen);
  end;
end;

function GetSisHandleFile;
begin
  Result := HFILE_ERROR;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := HANDLE_FILE;
end;

function GetSisVersions;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := MakeSisVersions(SIS_FILE_HEADER.MajorVersion, SIS_FILE_HEADER.MinorVersion);
end;

function GetSisInstalationDrive;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.InstallationDrive;
end;

function GetSisUID1;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.UID1;
end;

function GetSisUID2;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.UID2;
end;

function GetSisUID3;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.UID3;
end;

function GetSisUID4;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.UID4;
end;

function GetSisCountRequisite;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.NumberOfRequisites;  
end;

function GetSisRequisite;
var
  lpNumberOfBytesRead: DWORD;
begin
  Result := NIL;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfRequisites) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpRequisites + Index*SIZEOF(_REQUISITE),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, REQUISITE, SIZEOF(_REQUISITE), lpNumberOfBytesRead, NIL);
    Result := @REQUISITE;
  end;
end;

function GetSisRequisiteStr;
var
  Req: PTRequisite;
begin
  Req := GetSisRequisite(hSIS, Index);
  if (Req<>NIL) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    SetFilePointer(HANDLE_FILE, Req^.lpName, NIL, FILE_BEGIN);
    Result := ReadUnicode(HANDLE_FILE, Req^.NameLen);
  end;
end;

function GetSisCountCapabilities;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.NumberOfCapabilities;
end;

function GetSisCapabilities;
var
  lpNumberOfBytesRead: DWORD;
begin
  Result := NIL;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfCapabilities) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpCapabilities + Index*SIZEOF(_CAPABILITIES),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, CAPABILITIES, SIZEOF(_CAPABILITIES), lpNumberOfBytesRead, NIL);
    Result := @CAPABILITIES;
  end;
end;

function GetSisCountLanguages;
begin
  Result := 0;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
    Result := SIS_FILE_HEADER.NumberOfLanguages;
end;

function GetSisLanguage;
var
  lpNumberOfBytesRead: DWORD;
  lpLang: TLanguage;
begin
  Result := NIL;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  if (index<SIS_FILE_HEADER.NumberOfLanguages) then
  begin
    SetFilePointer(HANDLE_FILE,
      SIS_FILE_HEADER.lpLanguages + Index*SIZEOF(WORD),
      NIL, FILE_BEGIN);
    ReadFile(HANDLE_FILE, lpLang, SIZEOF(TLanguage), lpNumberOfBytesRead, NIL);
    Result := @lpLang;
  end;
end;

function GetSisLanguageStr;
begin
  Result := _LANGUAGES_STRINGS[GetSisLanguage(hSIS, Index)^.ID];
end;

function GetSisLanguageStrCode;
begin
  Result := _LANGUAGES_CODES[GetSisLanguage(hSIS, Index)^.ID];
end;

function ExtractSis;
var
  i: UINT;
  hFout: HFILE;
  lpNumberOfBytesRead, lpNumberOfBytesWrite: DWORD;
  lpFileName, lpExtractPath, lpExtractName, lpTmp: String;
  lpOpenBuff: _OFSTRUCT;
  inpBuf, outBuf: Pointer;
  outBytes: Integer;
begin
  Result := FALSE;
  if (hSIS <> INVALID_SIS_HANDLE) then
  with p_DATA_HANDLE_SIS(hSIS)^ do
  begin
    if (Assigned(PROCESS_EVENT)) then
    begin
      SetString(lpFileName, NIL, MAX_PATH);
      GetModuleFileName(HANDLE_FILE, LPCSTR(lpFileName), MAX_PATH);
      PROCESS_EVENT(hSIS, SIS_PROC_TYPE_EXTRACT, Integer(lpFileName), 0);
    end;

    for i:= 1 to SIS_FILE_HEADER.NumberOfFiles do
    begin
      ZeroMemory(@FILE_HEADER, SIZEOF(_FILE_HEADER));
      SetFilePointer(HANDLE_FILE, SIS_FILE_HEADER.lpFiles + (i-1)*SIZEOF(_FILE_HEADER), NIL, FILE_BEGIN);
      ReadFile(HANDLE_FILE, FILE_HEADER, SIZEOF(_FILE_HEADER), lpNumberOfBytesRead, NIL);
      if (FILE_HEADER.DestNameLen > 0) then
      begin
        SetFilePointer(HANDLE_FILE, FILE_HEADER.lpDestName+6, NIL, FILE_BEGIN);
        lpTmp := ReadUnicode(HANDLE_FILE, FILE_HEADER.DestNameLen-6);
        case Flags of
          SIS_SAVEDIR:lpFileName:=lpTmp;
          else
          begin
            lpFileName := ExtractFileName(lpTmp);
            if (length(lpFileName)=0) then lpFileName := lpTmp;
          end;
        end;
      end else
      begin
        SetFilePointer(HANDLE_FILE, FILE_HEADER.lpSrcName+6, NIL, FILE_BEGIN);
        lpTmp := ReadUnicode(HANDLE_FILE, FILE_HEADER.SrcNameLen-6);
        lpFileName := ExtractFileName(lpTmp);
      end;
      lpExtractName := lpDir + lpFileName;
      lpExtractPath := ExtractFilePath(lpExtractName);
      ForceDirectories(lpExtractPath);
      DeleteFile(LPCSTR(lpExtractName));
      hFout := OpenFile(LPCSTR(lpExtractName), lpOpenBuff, OF_CREATE);
      SetFilePointer(HANDLE_FILE, FILE_HEADER.lpFile, NIL, FILE_BEGIN);
      GetMem(inpBuf, FILE_HEADER.FileSize);
      ReadFile(HANDLE_FILE, inpBuf^, FILE_HEADER.FileSize, lpNumberOfBytesRead, NIL);
      zlib.DecompressBuf(inpBuf, FILE_HEADER.FileSize, 5, outBuf, outBytes);
      WriteFile(hFout, outBuf^, outBytes, lpNumberOfBytesWrite, NIL);
      FreeMem(outBuf);
      CloseHandle(hFout);
      FreeMem(inpBuf);
      DoSisEvent(hSIS, SIS_PROC_TYPE_FILE, Integer(@lpExtractName), 0, PROCESS_EVENT);
    end;
    Result := TRUE;
  end;      
end;

end.
